# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import io
import unittest

from trac.attachment import Attachment
from trac.mimeview.api import RenderingContext
from trac.resource import Resource
from trac.search.web_ui import SearchModule
from trac.test import MockPerm, mkdtemp
from trac.web.href import Href
from trac.wiki.tests import formatter


SEARCH_TEST_CASES = u"""
============================== search: link resolver
search:foo
search:"foo bar"
[search:bar Bar]
[search:bar]
[search:]
------------------------------
<p>
<a class="search" href="/search?q=foo">search:foo</a>
<a class="search" href="/search?q=foo+bar">search:"foo bar"</a>
<a class="search" href="/search?q=bar">Bar</a>
<a class="search" href="/search?q=bar">bar</a>
<a class="search" href="/search">search</a>
</p>
------------------------------
============================== search: link resolver with query arguments
search:foo?wiki=on
search:?q=foo&wiki=on
search:"foo bar?wiki=on"
search:"?q=foo bar&wiki=on"
[search:bar?ticket=on Bar in Tickets]
[search:?q=bar&ticket=on Bar in Tickets]
------------------------------
<p>
<a class="search" href="/search?q=foo&amp;wiki=on">search:foo?wiki=on</a>
<a class="search" href="/search?q=foo&amp;wiki=on">search:?q=foo&amp;wiki=on</a>
<a class="search" href="/search?q=foo+bar&amp;wiki=on">search:"foo bar?wiki=on"</a>
<a class="search" href="/search?q=foo+bar&amp;wiki=on">search:"?q=foo bar&amp;wiki=on"</a>
<a class="search" href="/search?q=bar&amp;ticket=on">Bar in Tickets</a>
<a class="search" href="/search?q=bar&amp;ticket=on">Bar in Tickets</a>
</p>
------------------------------
"""

ATTACHMENT_TEST_CASES = u"""
============================== attachment: link resolver (deprecated)
attachment:wiki:WikiStart:file.txt (deprecated)
attachment:ticket:123:file.txt (deprecated)
[attachment:wiki:WikiStart:file.txt file.txt] (deprecated)
[attachment:ticket:123:file.txt] (deprecated)
------------------------------
<p>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">attachment:wiki:WikiStart:file.txt</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a> (deprecated)
<a class="attachment" href="/attachment/ticket/123/file.txt" title="Attachment 'file.txt' in Ticket #123">attachment:ticket:123:file.txt</a><a class="trac-rawlink" href="/raw-attachment/ticket/123/file.txt" title="Download"></a> (deprecated)
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">file.txt</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a> (deprecated)
<a class="attachment" href="/attachment/ticket/123/file.txt" title="Attachment 'file.txt' in Ticket #123">ticket:123:file.txt</a><a class="trac-rawlink" href="/raw-attachment/ticket/123/file.txt" title="Download"></a> (deprecated)
</p>
------------------------------
============================== attachment: "foreign" links
attachment:file.txt:wiki:WikiStart
attachment:file.txt:ticket:123
[attachment:file.txt:wiki:WikiStart file.txt]
[attachment:file.txt:ticket:123]
attachment:foo.txt:wiki:SomePage/SubPage
------------------------------
<p>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">attachment:file.txt:wiki:WikiStart</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a>
<a class="attachment" href="/attachment/ticket/123/file.txt" title="Attachment 'file.txt' in Ticket #123">attachment:file.txt:ticket:123</a><a class="trac-rawlink" href="/raw-attachment/ticket/123/file.txt" title="Download"></a>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">file.txt</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a>
<a class="attachment" href="/attachment/ticket/123/file.txt" title="Attachment 'file.txt' in Ticket #123">file.txt:ticket:123</a><a class="trac-rawlink" href="/raw-attachment/ticket/123/file.txt" title="Download"></a>
<a class="attachment" href="/attachment/wiki/SomePage/SubPage/foo.txt" title="Attachment 'foo.txt' in SomePage/SubPage">attachment:foo.txt:wiki:SomePage/SubPage</a><a class="trac-rawlink" href="/raw-attachment/wiki/SomePage/SubPage/foo.txt" title="Download"></a>
</p>
------------------------------
============================== attachment: "local" links
attachment:file.txt
[attachment:file.txt that file]
------------------------------
<p>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">attachment:file.txt</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">that file</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt" title="Download"></a>
</p>
------------------------------
============================== attachment: "missing" links
attachment:foo.txt
[attachment:foo.txt other file]
------------------------------
<p>
<a class="missing attachment">attachment:foo.txt</a>
<a class="missing attachment">other file</a>
</p>
------------------------------
============================== attachment: "raw" links
raw-attachment:file.txt
[raw-attachment:file.txt that file]
------------------------------
<p>
<a class="attachment" href="/raw-attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">raw-attachment:file.txt</a>
<a class="attachment" href="/raw-attachment/wiki/WikiStart/file.txt" title="Attachment 'file.txt' in WikiStart">that file</a>
</p>
------------------------------
============================== attachment: raw format as explicit argument
attachment:file.txt?format=raw
[attachment:file.txt?format=raw that file]
------------------------------
<p>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt?format=raw" title="Attachment 'file.txt' in WikiStart">attachment:file.txt?format=raw</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt?format=raw" title="Download"></a>
<a class="attachment" href="/attachment/wiki/WikiStart/file.txt?format=raw" title="Attachment 'file.txt' in WikiStart">that file</a><a class="trac-rawlink" href="/raw-attachment/wiki/WikiStart/file.txt?format=raw" title="Download"></a>
</p>
------------------------------
""" # "

def attachment_setup(tc):
    import trac.ticket.api
    import trac.wiki.api
    tc.env.path = mkdtemp()
    with tc.env.db_transaction as db:
        db("INSERT INTO wiki (name,version) VALUES ('SomePage/SubPage',1)")
        db("INSERT INTO ticket (id) VALUES (123)")
    attachment = Attachment(tc.env, 'ticket', 123)
    attachment.insert('file.txt', io.BytesIO(b''), 0)
    attachment = Attachment(tc.env, 'wiki', 'WikiStart')
    attachment.insert('file.txt', io.BytesIO(b''), 0)
    attachment = Attachment(tc.env, 'wiki', 'SomePage/SubPage')
    attachment.insert('foo.txt', io.BytesIO(b''), 0)

def attachment_teardown(tc):
    tc.env.reset_db_and_disk()


EMAIL_TEST_CASE_DEFAULT = u"""
============================== mailto: obfuscated by default, like plain email
user@example.org vs. mailto:user@example.org
and [mailto:user@example.org Joe User]
------------------------------
<p>
user@\u2026 vs. mailto:user@\u2026
and Joe User
</p>
------------------------------
"""

def email_default_context():
    class NoEmailViewPerm(MockPerm):
        def has_permission(self, action, realm_or_resource=None, id=False,
                           version=False):
            return action != 'EMAIL_VIEW'
        __contains__ = has_permission

    context = RenderingContext(Resource('wiki', 'WikiStart'), href=Href('/'),
                               perm=NoEmailViewPerm())
    context.req = None # 1.0 FIXME .req shouldn't be required by formatter
    return context


EMAIL_TEST_CASE_NEVER_OBFUSCATE = u"""
============================== mailto: not obfuscated, unlike plain email
user@example.org vs. mailto:user@example.org
and [mailto:user@example.org Joe User]
------------------------------
<p>
user@\u2026 vs. <a class="mail-link" href="mailto:user@example.org"><span class="icon"></span>mailto:user@example.org</a>
and <a class="mail-link" href="mailto:user@example.org"><span class="icon"></span>Joe User</a>
</p>
------------------------------
"""

def email_never_obfuscate_setup(tc):
    tc.env.config.set('trac', 'never_obfuscate_mailto', True)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(formatter.test_suite(SEARCH_TEST_CASES, file=__file__))
    suite.addTest(formatter.test_suite(ATTACHMENT_TEST_CASES, file=__file__,
                                       context=('wiki', 'WikiStart'),
                                       setup=attachment_setup,
                                       teardown=attachment_teardown))
    suite.addTest(formatter.test_suite(EMAIL_TEST_CASE_DEFAULT, file=__file__,
                                       context=email_default_context()))
    suite.addTest(formatter.test_suite(EMAIL_TEST_CASE_NEVER_OBFUSCATE,
                                       file=__file__,
                                       context=email_default_context(),
                                       setup=email_never_obfuscate_setup))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import os
import unittest

from trac.mimeview.api import Mimeview
from trac.test import EnvironmentStub, MockPerm, MockRequest
from trac.ticket.test import insert_ticket
from trac.ticket.web_ui import TicketModule


class TicketConversionTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub()
        self.env.config.set('trac', 'templates_dir',
                            os.path.join(os.path.dirname(self.env.path),
                                         'templates'))
        self.ticket_module = TicketModule(self.env)
        self.mimeview = Mimeview(self.env)
        self.req = MockRequest(self.env, authname='anonymous')

    def tearDown(self):
        self.env.reset_db()

    def _create_a_ticket(self):
        return insert_ticket(self.env, owner='', reporter='santa',
                             summary='Foo', description='Bar',
                             foo='This is a custom field')

    def _create_a_ticket_with_email(self):
        return insert_ticket(self.env, owner='joe@example.org',
                             reporter='santa@example.org',
                             cc='cc1, cc2@example.org', summary='Foo',
                             description='Bar')

    def test_conversions(self):
        conversions = self.mimeview.get_supported_conversions(
            'trac.ticket.Ticket')
        expected = [('rss', 'RSS Feed', 'xml',
                     'trac.ticket.Ticket', 'application/rss+xml', 8,
                     self.ticket_module),
                    ('csv', 'Comma-delimited Text', 'csv',
                     'trac.ticket.Ticket', 'text/csv', 8,
                     self.ticket_module),
                    ('tab', 'Tab-delimited Text', 'tsv',
                     'trac.ticket.Ticket', 'text/tab-separated-values', 8,
                     self.ticket_module)]
        self.assertEqual(expected, list(conversions))

    def test_csv_conversion(self):
        ticket = self._create_a_ticket()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'csv')
        self.assertEqual(('\xef\xbb\xbf'
                          'id,summary,reporter,owner,description,status,'
                          'keywords,cc\r\n1,Foo,santa,,Bar,new,,\r\n',
                          'text/csv;charset=utf-8', 'csv'), csv)

    def test_csv_conversion_with_obfuscation(self):
        ticket = self._create_a_ticket_with_email()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'csv')
        self.assertEqual(
            ('\xef\xbb\xbf'
             'id,summary,reporter,owner,description,status,keywords,cc\r\n'
             '1,Foo,santa@…,joe@…,Bar,new,,cc1 cc2@…\r\n',
             'text/csv;charset=utf-8', 'csv'),
            csv)
        self.req.perm = MockPerm()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'csv')
        self.assertEqual(
            ('\xef\xbb\xbf'
             'id,summary,reporter,owner,description,status,keywords,cc\r\n'
             '1,Foo,santa@example.org,joe@example.org,Bar,new,,'
             'cc1 cc2@example.org\r\n',
             'text/csv;charset=utf-8', 'csv'),
            csv)

    def test_tab_conversion(self):
        ticket = self._create_a_ticket()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'tab')
        self.assertEqual(('\xef\xbb\xbf'
                          'id\tsummary\treporter\towner\tdescription\tstatus\t'
                          'keywords\tcc\r\n1\tFoo\tsanta\t\tBar\tnew\t\t\r\n',
                          'text/tab-separated-values;charset=utf-8', 'tsv'),
                         csv)

    def test_tab_conversion_with_obfuscation(self):
        ticket = self._create_a_ticket_with_email()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'tab')
        self.assertEqual(
            ('\xef\xbb\xbf'
             'id\tsummary\treporter\towner\tdescription\tstatus\tkeywords\t'
             'cc\r\n'
             '1\tFoo\tsanta@…\tjoe@…\tBar\tnew\t\tcc1 cc2@…\r\n',
             'text/tab-separated-values;charset=utf-8', 'tsv'),
            csv)
        self.req.perm = MockPerm()
        csv = self.mimeview.convert_content(self.req, 'trac.ticket.Ticket',
                                            ticket, 'tab')
        self.assertEqual(
            ('\xef\xbb\xbf'
             'id\tsummary\treporter\towner\tdescription\tstatus\tkeywords\t'
             'cc\r\n'
             '1\tFoo\tsanta@example.org\tjoe@example.org\tBar\tnew\t\t'
             'cc1 cc2@example.org\r\n',
             'text/tab-separated-values;charset=utf-8', 'tsv'),
            csv)

    def test_rss_conversion(self):
        ticket = self._create_a_ticket()
        content, mimetype, ext = self.mimeview.convert_content(
            self.req, 'trac.ticket.Ticket', ticket, 'rss')
        self.maxDiff = None
        self.assertEqual(("""<?xml version="1.0"?>



<rss version="2.0" xmlns:dc="http://purl.org/dc/elements/1.1/">

  <channel>
    <title>My Project: Ticket #1: Foo</title>
    <link>http://example.org/trac.cgi/ticket/1</link>
    <description>&lt;p&gt;
Bar
&lt;/p&gt;
</description>
    <language>en-us</language>
    <image>
      <title>My Project</title>
      <url>http://example.org/trac.cgi/chrome/site/your_project_logo.png</url>
      <link>http://example.org/trac.cgi/ticket/1</link>
    </image>
    <generator>Trac %s</generator>

 </channel>
</rss>""" % self.env.trac_version,
                          'application/rss+xml', 'xml'),
                         (content.replace('\r', ''), mimetype, ext))


def test_suite():
    return unittest.makeSuite(TicketConversionTestCase)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import unittest

from trac.wiki.tests import formatter

TEST_CASES = """
============================== htdocs: links resolver
htdocs:release-1.0.tar.gz

[htdocs:release-1.0.tar.gz Release 1.0]
------------------------------
<p>
<a href="/chrome/site/release-1.0.tar.gz">htdocs:release-1.0.tar.gz</a>
</p>
<p>
<a href="/chrome/site/release-1.0.tar.gz">Release 1.0</a>
</p>
------------------------------
"""

def test_suite():
    return formatter.test_suite(TEST_CASES, file=__file__)

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import unittest

from trac.core import ComponentManager
from trac.perm import PermissionSystem
from trac.resource import Resource, ResourceNotFound, render_resource_link
from trac.test import EnvironmentStub, MockRequest
from trac.ticket.roadmap import (
    DefaultTicketGroupStatsProvider, Milestone, MilestoneModule,
    RoadmapModule, TicketGroupStats, get_tickets_for_all_milestones,
    get_tickets_for_milestone)
from trac.ticket.test import insert_ticket
from trac.util.datefmt import datetime_now, utc
from trac.web.api import HTTPBadRequest, RequestDone
from trac.web.chrome import web_context
from trac.web.tests.api import RequestHandlerPermissionsTestCaseBase


class TicketGroupStatsTestCase(unittest.TestCase):

    def setUp(self):
        self.stats = TicketGroupStats('title', 'units')

    def test_init(self):
        self.assertEqual('title', self.stats.title, 'title incorrect')
        self.assertEqual('units', self.stats.unit, 'unit incorrect')
        self.assertEqual(0, self.stats.count, 'count not zero')
        self.assertEqual(0, len(self.stats.intervals), 'intervals not empty')

    def test_add_iterval(self):
        self.stats.add_interval('intTitle', 3, {'k1': 'v1'}, 'css', 0)
        self.stats.refresh_calcs()
        self.assertEqual(3, self.stats.count, 'count not incremented')
        int = self.stats.intervals[0]
        self.assertEqual('intTitle', int['title'], 'title incorrect')
        self.assertEqual(3, int['count'], 'count incorrect')
        self.assertEqual({'k1': 'v1'}, int['qry_args'],
                         'query args incorrect')
        self.assertEqual('css', int['css_class'], 'css class incorrect')
        self.assertEqual(100, int['percent'], 'percent incorrect')
        self.stats.add_interval('intTitle', 3, {'k1': 'v1'}, 'css', 0)
        self.stats.refresh_calcs()
        self.assertEqual(50, int['percent'], 'percent not being updated')

    def test_add_interval_no_prog(self):
        self.stats.add_interval('intTitle', 3, {'k1': 'v1'}, 'css', 0)
        self.stats.add_interval('intTitle', 5, {'k1': 'v1'}, 'css', 0)
        self.stats.refresh_calcs()
        self.assertEqual(0, self.stats.done_count, 'count added for no prog')
        self.assertEqual(0, self.stats.done_percent, 'percent incremented')

    def test_add_interval_prog(self):
        self.stats.add_interval('intTitle', 3, {'k1': 'v1'}, 'css', 0)
        self.stats.add_interval('intTitle', 1, {'k1': 'v1'}, 'css', 1)
        self.stats.refresh_calcs()
        self.assertEqual(4, self.stats.count, 'count not incremented')
        self.assertEqual(1, self.stats.done_count, 'count not added to prog')
        self.assertEqual(25, self.stats.done_percent, 'done percent not incr')

    def test_add_interval_fudging(self):
        self.stats.add_interval('intTitle', 3, {'k1': 'v1'}, 'css', 0)
        self.stats.add_interval('intTitle', 5, {'k1': 'v1'}, 'css', 1)
        self.stats.refresh_calcs()
        self.assertEqual(8, self.stats.count, 'count not incremented')
        self.assertEqual(5, self.stats.done_count, 'count not added to prog')
        self.assertEqual(62, self.stats.done_percent,
                         'done percnt not fudged downward')
        self.assertEqual(62, self.stats.intervals[1]['percent'],
                         'interval percent not fudged downward')
        self.assertEqual(38, self.stats.intervals[0]['percent'],
                         'interval percent not fudged upward')


class DefaultTicketGroupStatsProviderTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True)

        self.milestone1 = Milestone(self.env)
        self.milestone1.name = 'Test'
        self.milestone1.insert()
        self.milestone2 = Milestone(self.env)
        self.milestone2.name = 'Test2'
        self.milestone2.insert()

        tkt1 = insert_ticket(self.env, summary='Foo', milestone='Test',
                             owner='foman', status='new')
        tkt2 = insert_ticket(self.env, summary='Bar', milestone='Test',
                             status='closed', owner='barman')
        tkt3 = insert_ticket(self.env, summary='Sum', milestone='Test',
                             owner='suman', status='reopened')
        self.tkt1 = tkt1
        self.tkt2 = tkt2
        self.tkt3 = tkt3

        prov = DefaultTicketGroupStatsProvider(ComponentManager())
        prov.env = self.env
        prov.config = self.env.config
        self.stats = prov.get_ticket_group_stats([tkt1.id, tkt2.id, tkt3.id])

    def tearDown(self):
        self.env.reset_db()

    def test_stats(self):
        self.assertEqual(self.stats.title, 'ticket status', 'title incorrect')
        self.assertEqual(self.stats.unit, 'tickets', 'unit incorrect')
        self.assertEqual(2, len(self.stats.intervals),
                         'more than 2 intervals')

    def test_closed_interval(self):
        closed = self.stats.intervals[0]
        self.assertEqual('closed', closed['title'], 'closed title incorrect')
        self.assertEqual('closed', closed['css_class'],
                         'closed class incorrect')
        self.assertTrue(closed['overall_completion'],
                        'closed should contribute to overall completion')
        self.assertEqual({'status': ['closed'], 'group': ['resolution']},
                         closed['qry_args'], 'qry_args incorrect')
        self.assertEqual(1, closed['count'], 'closed count incorrect')
        self.assertEqual(33, closed['percent'], 'closed percent incorrect')

    def test_open_interval(self):
        open = self.stats.intervals[1]
        self.assertEqual('active', open['title'], 'open title incorrect')
        self.assertEqual('open', open['css_class'], 'open class incorrect')
        self.assertFalse(open['overall_completion'],
                         "open shouldn't contribute to overall completion")
        self.assertEqual({'status':
                          [u'assigned', u'new', u'accepted', u'reopened']},
                         open['qry_args'], 'qry_args incorrect')
        self.assertEqual(2, open['count'], 'open count incorrect')
        self.assertEqual(67, open['percent'], 'open percent incorrect')


class MilestoneModuleTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True)
        self.mmodule = MilestoneModule(self.env)
        self.terms = ['MilestoneAlpha', 'MilestoneBeta', 'MilestoneGamma']
        for term in self.terms + [' '.join(self.terms)]:
            m = Milestone(self.env)
            m.name = term
            m.due = datetime_now(utc)
            m.description = u"""\
Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod \
tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim \
veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea \
commodo consequat. Duis aute irure dolor in reprehenderit in voluptate \
velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat \
cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id \
est laborum."""
            m.insert()

    def tearDown(self):
        self.env.reset_db()

    def test_invalid_post_request_raises_exception(self):
        req = MockRequest(self.env, method='POST', action=None,
                          args={'id': 'milestone1'})

        self.assertRaises(HTTPBadRequest,
                          MilestoneModule(self.env).process_request, req)

    def test_milestone_redirects_to_roadmap(self):
        """The path /milestone redirects to /roadmap."""
        def test_milestone_redirect(authname=None):
            req = MockRequest(self.env, method='GET', path_info='/milestone',
                              authname=authname)
            with self.assertRaises(RequestDone):
                self.mmodule.process_request(req)
            self.assertEqual('http://example.org/trac.cgi/roadmap',
                             req.headers_sent['Location'])
            self.assertEqual('302 Found', req._status)
            return req

        # Redirects for user with MILESTONE_VIEW
        req = test_milestone_redirect()
        self.assertIn('MILESTONE_VIEW', req.perm)

        # Redirects for user without MILESTONE_VIEW
        perm_sys = PermissionSystem(self.env)
        perm_sys.revoke_permission('anonymous', 'MILESTONE_VIEW')
        req = test_milestone_redirect('user1')
        self.assertNotIn('MILESTONE_VIEW', req.perm)

    def test_get_search_filters(self):
        req = MockRequest(self.env)
        filters = self.mmodule.get_search_filters(req)
        filters = list(filters)
        self.assertEqual(1, len(filters))
        self.assertEqual(2, len(filters[0]))
        self.assertEqual('milestone', filters[0][0])
        self.assertEqual('Milestones', filters[0][1])

    def test_get_search_results_milestone_not_in_filters(self):
        req = MockRequest(self.env)
        results = self.mmodule.get_search_results(req, self.terms, [])
        self.assertEqual([], list(results))

    def test_get_search_results_matches_all_terms(self):
        req = MockRequest(self.env)
        milestone = Milestone(self.env, ' '.join(self.terms))
        results = self.mmodule.get_search_results(req, self.terms,
                                                  ['milestone'])
        results = list(results)
        self.assertEqual(1, len(results))
        self.assertEqual(5, len(results[0]))
        self.assertEqual('/trac.cgi/milestone/' +
                         milestone.name.replace(' ', '%20'),
                         results[0][0])
        self.assertEqual('Milestone ' + milestone.name, results[0][1])
        self.assertEqual(milestone.due, results[0][2])
        self.assertEqual('', results[0][3])
        shorten_desc = u"""\
Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod \
tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, \
quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo \
consequat. Duis a ..."""
        self.assertEqual(shorten_desc, results[0][4])

    def test_default_group_by_default(self):
        """Default `default_group_by` is `component`."""
        req = MockRequest(self.env, args={'id': 'milestone1'})

        data = self.mmodule.process_request(req)[1]

        self.assertIn('grouped_by', data)
        self.assertEqual('component', data['grouped_by'])

    def test_default_group_by(self):
        """Option `default_group_by` is set in configuration."""
        self.env.config.set('milestone', 'default_group_by', 'priority')
        req = MockRequest(self.env, args={'id': 'milestone1'})

        data = self.mmodule.process_request(req)[1]

        self.assertIn('grouped_by', data)
        self.assertEqual('priority', data['grouped_by'])

    def test_get_search_results_matches_ignorecase(self):
        req = MockRequest(self.env)

        def search(terms):
            return list(self.mmodule.get_search_results(req, terms,
                                                        ['milestone']))

        results = search(self.terms)
        self.assertEqual(results, search([t.lower() for t in self.terms]))
        self.assertEqual(results, search([t.upper() for t in self.terms]))


class MilestoneModulePermissionsTestCase(RequestHandlerPermissionsTestCaseBase):

    def setUp(self):
        super(MilestoneModulePermissionsTestCase, self).setUp(MilestoneModule)

    def test_milestone_notfound_with_milestone_create(self):
        self.grant_perm('anonymous', 'MILESTONE_VIEW')
        self.grant_perm('anonymous', 'MILESTONE_CREATE')

        req = MockRequest(self.env, path_info='/milestone/milestone5')
        res = self.process_request(req)

        self.assertEqual('milestone_edit.html', res[0])
        self.assertEqual('milestone5', res[1]['milestone'].name)
        self.assertEqual("Milestone milestone5 does not exist. You can"
                         " create it here.", req.chrome['notices'][0])

    def test_milestone_notfound_without_milestone_create(self):
        self.grant_perm('anonymous', 'MILESTONE_VIEW')

        req = MockRequest(self.env, authname='anonymous',
                          path_info='/milestone/milestone5')

        self.assertRaises(ResourceNotFound, self.process_request, req)


class RoadmapTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub()
        self.env.config.set('ticket-custom', 'project', 'text')
        values = [
            ('Summary', 'new', 'milestone1', 'joe'),
            ('Summary', 'new', 'milestone2', 'joe'),
            ('Summary', 'new', '',           'joe'),
            ('Summary', 'new', None,         'john'),
            ('Summary', 'new', 'milestone1', 'john'),
            ('Summary', 'new', 'milestone2', 'blah'),
            ('Summary', 'new', '',           'blah'),
            ('Summary', 'new', None,         'blah'),
            ('Summary', 'new', 'milestone1', 'blah'),
            ('Summary', 'new', 'milestone2', 'blah'),
        ]
        custom_values = [
            (1, 'foo'), (2, 'foo'), (3, 'foo'), (4, 'foo'), (5, 'foo'),
            (6, 'bar'), (7, 'bar'), (8, 'baz'), (9, 'baz'), (10, 'baz'),
        ]
        with self.env.db_transaction as db:
            cursor = db.cursor()
            cursor.executemany("""
                INSERT INTO ticket (summary,status,milestone,owner)
                VALUES (%s,%s,%s,%s)
                """, values)
            cursor.executemany("""
                INSERT INTO ticket_custom (ticket,name,value)
                VALUES (%s,'project',%s)
                """, custom_values)

    def tearDown(self):
        self.env.reset_db()

    def insert_milestone(self, name, due=None):
        milestone = Milestone(self.env)
        milestone.name = name
        milestone.due = due
        milestone.insert()
        return milestone

    def test_get_tickets_for_all_milestones(self):
        tickets = get_tickets_for_all_milestones(self.env, field='owner')
        milestone1 = [{'id': 9, 'status': 'new', 'owner': 'blah'},
                      {'id': 1, 'status': 'new', 'owner': 'joe'},
                      {'id': 5, 'status': 'new', 'owner': 'john'}]
        milestone2 = [{'id': 6, 'status': 'new', 'owner': 'blah'},
                      {'id': 10, 'status': 'new', 'owner': 'blah'},
                      {'id': 2, 'status': 'new', 'owner': 'joe'}]
        self.assertEqual(milestone1, tickets['milestone1'])
        self.assertEqual(milestone1,
                         get_tickets_for_milestone(self.env,
                                                   milestone='milestone1',
                                                   field='owner'))
        self.assertEqual(milestone2, tickets['milestone2'])
        self.assertEqual(milestone2,
                         get_tickets_for_milestone(self.env,
                                                   milestone='milestone2',
                                                   field='owner'))
        self.assertEqual(['milestone1', 'milestone2'], sorted(tickets))

    def test_get_tickets_for_all_milestones_custom_field(self):
        tickets = get_tickets_for_all_milestones(self.env, field='project')
        milestone1 = [{'id': 9, 'status': 'new', 'project': 'baz'},
                      {'id': 1, 'status': 'new', 'project': 'foo'},
                      {'id': 5, 'status': 'new', 'project': 'foo'}]
        milestone2 = [{'id': 6, 'status': 'new', 'project': 'bar'},
                      {'id': 10, 'status': 'new', 'project': 'baz'},
                      {'id': 2, 'status': 'new', 'project': 'foo'}]
        self.assertEqual(milestone1, tickets['milestone1'])
        self.assertEqual(milestone1,
                         get_tickets_for_milestone(self.env,
                                                   milestone='milestone1',
                                                   field='project'))
        self.assertEqual(milestone2, tickets['milestone2'])
        self.assertEqual(milestone2,
                         get_tickets_for_milestone(self.env,
                                                   milestone='milestone2',
                                                   field='project'))
        self.assertEqual(['milestone1', 'milestone2'], sorted(tickets))

    def test_export_ical_from_roadmap(self):
        self.insert_milestone('milestone1', datetime_now(utc))
        self.insert_milestone('milestone2')
        rm = RoadmapModule(self.env)
        req = MockRequest(self.env, path_info='/roadmap',
                          args={'format': 'ics'})

        self.assertTrue(rm.match_request(req))
        with self.assertRaises(RequestDone):
            rm.process_request(req)

        self.assertEqual('200 Ok', req.status_sent[0])
        self.assertRegexpMatches(req.response_sent.getvalue(), """\
BEGIN:VCALENDAR
VERSION:2.0
PRODID:-//Edgewall Software//NONSGML Trac [\.\w]+//EN
METHOD:PUBLISH
X-WR-CALNAME:My Project - Roadmap
X-WR-CALDESC:My example project
X-WR-TIMEZONE:UTC
BEGIN:VEVENT
UID:</trac.cgi/milestone/milestone1@example.org/trac.cgi>
DTSTAMP:\w+
DTSTART;VALUE=DATE:\w+
SUMMARY:Milestone milestone1
URL:http://example.org/trac.cgi/milestone/milestone1
END:VEVENT
END:VCALENDAR
""".replace('\n', '\r\n'))


class ResourceTestCase(unittest.TestCase):
    """Test cases for milestone resources."""

    def setUp(self):
        self.env = EnvironmentStub()

    def test_resource_link_ticket_context_milestone_exists(self):
        """Resource link in ticket context for viewable milestone.
        """
        milestone = Milestone(self.env)
        milestone.name = 'milestone1'
        milestone.insert()
        req = MockRequest(self.env, path_info='/ticket/1')
        resource = Resource('milestone', 'milestone1')
        context = web_context(req)
        link = render_resource_link(self.env, context, resource,
                                    format='compact')
        self.assertEqual('<a class="milestone" href="/trac.cgi/milestone/'
                         'milestone1" title="No date set">milestone1</a>',
                         unicode(link))

    def test_resource_link_ticket_context_milestone_missing(self):
        """Resource link in ticket context for non-existent milestone.
        """
        req = MockRequest(self.env, path_info='/ticket/1', authname='user1')
        resource = Resource('milestone', 'milestone1')
        context = web_context(req)
        link = render_resource_link(self.env, context, resource,
                                    format='compact')
        self.assertEqual('<a class="milestone missing">milestone1</a>',
                         unicode(link))

    def test_resource_link_ticket_context_milestone_no_view_perm(self):
        """Resource link in ticket context with no milestone view permission.
        """
        milestone = Milestone(self.env)
        milestone.name = 'milestone1'
        milestone.insert()
        req = MockRequest(self.env, path_info='/ticket/1', authname='user1')
        resource = Resource('milestone', 'milestone1')
        context = web_context(req)
        link = render_resource_link(self.env, context, resource,
                                    format='compact')
        self.assertEqual('<a class="milestone">milestone1</a>',
                         unicode(link))


def in_tlist(ticket, list):
    return len([t for t in list if t['id'] == ticket.id]) > 0


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TicketGroupStatsTestCase))
    suite.addTest(unittest.makeSuite(DefaultTicketGroupStatsProviderTestCase))
    suite.addTest(unittest.makeSuite(MilestoneModuleTestCase))
    suite.addTest(unittest.makeSuite(MilestoneModulePermissionsTestCase))
    suite.addTest(unittest.makeSuite(RoadmapTestCase))
    suite.addTest(unittest.makeSuite(ResourceTestCase))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2004-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import datetime
import unittest

import trac.versioncontrol.admin
from trac.admin.console import TracAdmin
from trac.admin.test import TracAdminTestCaseBase
from trac.core import Component, implements
from trac.test import EnvironmentStub, Mock
from trac.util.datefmt import utc
from trac.versioncontrol.api import Changeset, DbRepositoryProvider, \
                                    IRepositoryConnector, NoSuchChangeset, \
                                    Repository


class MockRepositoryConnector(Component):

    implements(IRepositoryConnector)

    def get_supported_types(self):
        yield 'mock_type', 8

    def get_repository(self, repos_type, repos_dir, params):
        def get_changeset(rev):
            if rev == 'invalid':
                raise NoSuchChangeset(rev)
            return Mock(Changeset, repos, rev, 'message', 'author',
                        datetime.datetime(2001, 1, 1, tzinfo=utc))

        def get_node(path, rev):
            pass

        repos = Mock(Repository, params['name'], params, self.log,
                     get_youngest_rev=lambda: 1,
                     get_changeset=get_changeset,
                     get_node=get_node,
                     previous_rev=lambda rev, path='': None,
                     next_rev=lambda rev, path='': None)
        return repos


class TracAdminTestCase(TracAdminTestCaseBase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True, enable=('trac.*',),
                                   disable=('trac.tests.*',))
        self.admin = TracAdmin()
        self.admin.env_set('', self.env)
        provider = DbRepositoryProvider(self.env)
        provider.add_repository('mock', '/', 'mock_type')

    def tearDown(self):
        self.env = None

    def test_changeset_add_no_repository_revision(self):
        rv, output = self.execute('changeset added')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_add_no_revision(self):
        rv, output = self.execute('changeset added mock')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_modify_no_repository_revision(self):
        rv, output = self.execute('changeset modified')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_modify_no_revision(self):
        rv, output = self.execute('changeset modified mock')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_add_invalid_repository(self):
        rv, output = self.execute('changeset added invalid 123')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_modify_invalid_repository(self):
        rv, output = self.execute('changeset modified invalid 123')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_add_invalid_changeset(self):
        rv, output = self.execute('changeset added mock invalid')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)

    def test_changeset_modify_invalid_changeset(self):
        rv, output = self.execute('changeset modified mock invalid')
        self.assertEqual(2, rv, output)
        self.assertExpectedResult(output)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TracAdminTestCase))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import io
import os.path
import re
import sys

from contextlib import closing
from pkg_resources import resource_listdir, resource_string

from trac.loader import load_components
from trac.test import EnvironmentStub, Mock, MockPerm
from trac.util.text import printerr, printout
from trac.web.chrome import web_context
from trac.web.href import Href
from trac.wiki.formatter import format_to_html
from trac.wiki.model import WikiPage

try:
    import html2rest
except ImportError:
    printerr("The html2rest package must be installed.")
    sys.exit(1)


class Parser(html2rest.Parser):

    def __init__(self, writer=sys.stdout, encoding='utf8', relroot=None,
                 relpath=None):
        html2rest.Parser.__init__(self, writer, encoding, relroot, relpath)
        self.links = {}

    def end_a(self):
        if '#pending' in self.hrefs:
            href = self.hrefs['#pending']
            label = self.hrefs[href]
            key = label.lower()
            if key not in self.links:
                self.links[key] = (label, href)
            elif href != self.links[key][1]:
                alt = label
                while True:
                    alt += '*'
                    if alt not in self.links:
                        break
                    continue
                self.data(alt[len(label):])
                self.hrefs[href] = alt
                self.links[alt] = (alt, href)
            self.data('`_')
            del self.hrefs['#pending']

    def end_body(self):
        self.end_p()
        for label, href in self.links.itervalues():
            if href[0] != '#':
                self.writeline('.. _%s: %s' % (label, href))
        self.end_p()


def wiki2rest(env, context, wiki):
    text = re.sub('\r?\n', '\n', wiki.text)
    text = re.sub(r'\[\[TracGuideToc\]\]\r?\n?', '', text)
    text = re.sub(r'\[\[PageOutline\([^\)]*\)\]\]\r?\n?', '', text)
    html = unicode(format_to_html(env, context, text))
    html = html.replace(u'<span class="icon">\u200b</span>', '')
    html = re.sub(r'<em>\s*([^<]*?)\s*</em>', r'<em>\1</em>', html)
    html = '<html><body>%s</body></html>' % html
    writer = io.BytesIO()
    with closing(Parser(writer, 'utf-8', None, None)) as parser:
        parser.feed(html)
    rst = writer.getvalue().strip('\n')
    rst = re.sub('\n{4,}', '\n\n\n', rst)
    # sort links
    rst = re.sub(r'(?:\n\.\. _[^\n]*)+\Z',
                 lambda m: '\n'.join(sorted(m.group(0).split('\n'),
                                            key=lambda v: v.lower())),
                 rst)
    if any(ord(c) > 0x7f for c in rst):
        # Trac detects utf-8 using BOM
        rst = '%s.. charset=utf-8\n\n%s' % (codecs.BOM_UTF8, rst)
    return rst + '\n'


def main():
    names = sorted(name for name in resource_listdir('trac.wiki',
                                                     'default-pages')
                        if not name.startswith('.'))

    env = EnvironmentStub()
    load_components(env)
    with env.db_transaction:
        for name in names:
            wiki = WikiPage(env, name)
            wiki.text = resource_string('trac.wiki', 'default-pages/' +
                                        name).decode('utf-8')
            if wiki.text:
                wiki.save('trac', '')
            else:
                printout('%s: Skipped empty page' % name)

    req = Mock(href=Href('/'), abs_href=Href('https://trac.edgewall.org/'),
               perm=MockPerm(), chrome={})
    for name in sys.argv[1:]:
        name = os.path.basename(name)
        wiki = WikiPage(env, name)
        if not wiki.exists:
            continue
        context = web_context(req, wiki.resource, absurls=True)
        rst = wiki2rest(env, context, wiki)
        sys.stdout.write(rst)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

from __future__ import absolute_import

import unittest
try:
    import pygments
    have_pygments = True
except ImportError:
    have_pygments = False

from trac.tests.functional import FunctionalTwillTestCaseSetup, tc


class TestSyntaxHighlightingPreferences(FunctionalTwillTestCaseSetup):
    def runTest(self):
        """Set preferences for syntax highlighting."""
        self._tester.go_to_preferences("Syntax Highlighting")
        tc.find('<option selected="selected" value="trac">')
        tc.formvalue('userprefs', 'style', 'Emacs')
        tc.submit()
        tc.find('<option selected="selected" value="emacs">')


def functionalSuite(suite=None):
    if not suite:
        import trac.tests.functional
        suite = trac.tests.functional.functionalSuite()
    if have_pygments:
        suite.addTest(TestSyntaxHighlightingPreferences())
    return suite


test_suite = functionalSuite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2005-2023 Edgewall Software
# Copyright (C) 2005-2006 Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.
#
# Include a basic SMTP server, based on L. Smithson
# (lsmithson@open-networks.co.uk) extensible Python SMTP Server
#

import base64
import io
import quopri
import re
import unittest
from datetime import datetime, timedelta

from trac.attachment import Attachment
from trac.notification.api import NotificationSystem
from trac.perm import PermissionSystem
from trac.test import EnvironmentStub, MockRequest, mkdtemp
from trac.tests.notification import SMTP_TEST_PORT, SMTPThreadedServer, \
                                    parse_smtp_message
from trac.ticket.model import Ticket
from trac.ticket.notification import (
    BatchTicketChangeEvent, Subscription, TicketChangeEvent,
    TicketNotificationSystem)
from trac.ticket.test import insert_ticket
from trac.ticket.web_ui import TicketModule
from trac.util.datefmt import datetime_now, utc

MAXBODYWIDTH = 76
smtpd = None


def setUpModule():
    global smtpd
    smtpd = SMTPThreadedServer(SMTP_TEST_PORT)
    smtpd.start()


def tearDownModule():
    smtpd.stop()


def notify_ticket_created(env, ticket):
    smtpd.cleanup()
    event = TicketChangeEvent('created', ticket, ticket['time'],
                              ticket['reporter'])
    NotificationSystem(env).notify(event)


def notify_ticket_changed(env, ticket, author='anonymous'):
    smtpd.cleanup()
    event = TicketChangeEvent('changed', ticket, ticket['changetime'], author)
    NotificationSystem(env).notify(event)


def config_subscriber(env, updater=False, owner=False, reporter=False):
    section = 'notification-subscriber'
    env.config.set(section, 'always_notify_cc', 'CarbonCopySubscriber')
    if updater:
        env.config.set(section, 'always_notify_updater',
                       'TicketUpdaterSubscriber')
        env.config.set(section, 'always_notify_previous_updater',
                       'TicketPreviousUpdatersSubscriber')
    if owner:
        env.config.set(section, 'always_notify_owner',
                       'TicketOwnerSubscriber')
    if reporter:
        env.config.set(section, 'always_notify_reporter',
                       'TicketReporterSubscriber')
    del NotificationSystem(env).subscriber_defaults


def config_smtp(env):
    env.config.set('project', 'name', 'TracTest')
    env.config.set('notification', 'smtp_enabled', 'true')
    env.config.set('notification', 'smtp_port', str(SMTP_TEST_PORT))
    env.config.set('notification', 'smtp_server', smtpd.host)
    # Note: when specifying 'localhost', the connection may be attempted
    #       for '::1' first, then only '127.0.0.1' after a 1s timeout


def extract_recipients(header):
    rcpts = [rcpt.strip() for rcpt in header.split(',')]
    return [rcpt for rcpt in rcpts if rcpt]


class RecipientTestCase(unittest.TestCase):
    """Notification test cases for email recipients."""

    def setUp(self):
        self.env = EnvironmentStub(default_data=True)
        config_smtp(self.env)

    def tearDown(self):
        smtpd.cleanup()
        self.env.reset_db()

    def _insert_ticket(self, **props):
        return insert_ticket(self.env, **props)

    def test_no_recipients(self):
        """No recipient case"""
        ticket = insert_ticket(self.env, reporter='anonymous', summary='Foo')
        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()
        sender = smtpd.get_sender()
        message = smtpd.get_message()
        self.assertEqual(0, len(recipients))
        self.assertIsNone(sender)
        self.assertIsNone(message)

    def _test_smtp_always_cc(self, key, sep):
        cc_list = ('joe.user@example.net', 'joe.bar@example.net')
        self.env.config.set('notification', key, sep.join(cc_list))
        ticket = self._insert_ticket(reporter='joe.bar@example.org',
                                     owner='joe.user@example.net',
                                     summary='New ticket recipients')

        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()

        self.assertEqual(2, len(recipients))
        for r in cc_list:
            self.assertIn(r, recipients)

    def test_smtp_always_cc_comma_separator(self):
        self._test_smtp_always_cc('smtp_always_cc', ', ')

    def test_smtp_always_cc_space_separator(self):
        self._test_smtp_always_cc('smtp_always_cc', ' ')

    def test_smtp_always_bcc_comma_separator(self):
        self._test_smtp_always_cc('smtp_always_bcc', ', ')

    def test_smtp_always_bcc_space_separator(self):
        self._test_smtp_always_cc('smtp_always_bcc', ' ')

    def test_cc_permission_group_new_ticket(self):
        """Permission groups are resolved in CC for new ticket."""
        config_subscriber(self.env)
        ticket = self._insert_ticket(reporter='joe.bar@example.org',
                                     owner='joe.user@example.net',
                                     cc='group1, user1@example.net',
                                     summary='CC permission group')
        group1 = ('user2@example.net', 'user3@example.net')
        perm = PermissionSystem(self.env)
        for user in group1:
            perm.grant_permission(user, 'group1')

        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()

        self.assertEqual(3, len(recipients))
        for user in group1 + ('user1@example.net',):
            self.assertIn(user, recipients)

    def test_cc_permission_group_changed_ticket(self):
        """Permission groups are resolved in CC for ticket change."""
        config_subscriber(self.env)
        ticket = self._insert_ticket(reporter='joe.bar@example.org',
                                     owner='joe.user@example.net',
                                     cc='user1@example.net',
                                     summary='CC permission group')
        group1 = ('user2@example.net',)
        perm = PermissionSystem(self.env)
        for user in group1:
            perm.grant_permission(user, 'group1')

        ticket['cc'] += ', group1'
        ticket.save_changes('joe.bar2@example.com', 'This is a change')
        notify_ticket_changed(self.env, ticket)
        recipients = smtpd.get_recipients()

        self.assertEqual(2, len(recipients))
        for user in group1 + ('user1@example.net',):
            self.assertIn(user, recipients)

    def test_all_subscribers_permission_group(self):
        """Test for regression of https://trac.edgewall.org/ticket/13515"""

        config_subscriber(self.env, reporter=True, owner=True, updater=True)
        self.env.insert_users([
            ('u-reporter', 'User Reporter', 'reporter@example.org', 1),
            ('u-owner', 'User Owner', 'owner@example.org', 1),
            ('u-updater', 'User Updater', 'updater@example.org', 1),
            ('u-cc', 'User Cc', 'cc@example.org', 1),
        ])
        perm = PermissionSystem(self.env)
        for name in ('reporter', 'owner', 'updater', 'cc'):
            group = 'u-%s' % name
            perm.grant_permission('%s-1@example.net' % name, group)
            perm.grant_permission('%s-2@example.net' % name, group)

        ticket = self._insert_ticket(reporter='u-reporter', owner='u-owner',
                                     cc='u-cc',
                                     summary='Subscribers w/ permission group')
        notify_ticket_created(self.env, ticket)
        self.assertEqual(['cc-1@example.net', 'cc-2@example.net',
                          'owner@example.org', 'reporter@example.org',
                         ], sorted(smtpd.get_recipients()))

        ticket.save_changes('u-updater', 'This is a change')
        notify_ticket_changed(self.env, ticket)
        self.assertEqual(['cc-1@example.net', 'cc-2@example.net',
                          'owner@example.org', 'reporter@example.org',
                          'updater@example.org',
                         ], sorted(smtpd.get_recipients()))

    def test_new_ticket_recipients(self):
        """Report and CC list should be in recipient list for new tickets."""
        config_subscriber(self.env, updater=True)
        always_cc = ('joe.user@example.net', 'joe.bar@example.net')
        ticket_cc = ('joe.user@example.com', 'joe.bar@example.org')
        self.env.config.set('notification', 'smtp_always_cc',
                            ', '.join(always_cc))
        ticket = insert_ticket(self.env, reporter='joe.bar@example.org',
                               owner='joe.user@example.net',
                               cc=' '.join(ticket_cc),
                               summary='New ticket recipients')
        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()
        for r in always_cc + ticket_cc + \
                (ticket['owner'], ticket['reporter']):
            self.assertIn(r, recipients)

    def test_cc_only(self):
        """Notification w/o explicit recipients but Cc: (#3101)"""
        always_cc = ('joe.user@example.net', 'joe.bar@example.net')
        self.env.config.set('notification', 'smtp_always_cc',
                            ', '.join(always_cc))
        ticket = insert_ticket(self.env, summary='Foo')
        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()
        for r in always_cc:
            self.assertIn(r, recipients)

    def test_always_notify_updater(self):
        """The `always_notify_updater` option."""
        def _test_updater(enabled):
            config_subscriber(self.env, updater=enabled)
            ticket = insert_ticket(self.env, reporter='joe.user@example.org',
                                   summary=u'This is a súmmäry')
            now = datetime_now(utc)
            ticket.save_changes('joe.bar2@example.com', 'This is a change',
                                when=now)
            notify_ticket_changed(self.env, ticket)
            recipients = smtpd.get_recipients()
            if enabled:
                self.assertEqual(1, len(recipients))
                self.assertIn('joe.bar2@example.com', recipients)
            else:
                self.assertEqual(0, len(recipients))
                self.assertNotIn('joe.bar2@example.com', recipients)

        # Validate with and without a default domain
        for enable in False, True:
            _test_updater(enable)

    def test_always_notify_owner(self):
        """The `always_notify_owner` option."""
        def _test_reporter(enabled):
            config_subscriber(self.env, owner=enabled)
            ticket = insert_ticket(self.env, summary='Foo',
                                   reporter=u'joe@example.org',
                                   owner=u'jim@example.org')
            now = datetime_now(utc)
            ticket.save_changes('joe@example.org', 'this is my comment',
                                when=now)
            notify_ticket_changed(self.env, ticket)
            recipients = smtpd.get_recipients()
            if enabled:
                self.assertEqual(1, len(recipients))
                self.assertEqual('jim@example.org', recipients[0])
            else:
                self.assertEqual(0, len(recipients))

        for enable in False, True:
            _test_reporter(enable)

    def test_always_notify_reporter(self):
        """Notification to reporter w/ updater option disabled (#3780)"""
        def _test_reporter(enabled):
            config_subscriber(self.env, reporter=enabled)
            ticket = insert_ticket(self.env, summary='Foo',
                                   reporter=u'joe@example.org')
            now = datetime_now(utc)
            ticket.save_changes('joe@example.org', 'this is my comment',
                                when=now)
            notify_ticket_changed(self.env, ticket)
            recipients = smtpd.get_recipients()
            if enabled:
                self.assertEqual(1, len(recipients))
                self.assertEqual('joe@example.org', recipients[0])
            else:
                self.assertEqual(0, len(recipients))

        for enable in False, True:
            _test_reporter(enable)

    def test_notify_new_tickets(self):
        """Notification to NewTicketSubscribers."""
        def _test_new_ticket():
            ticket = Ticket(self.env)
            ticket['reporter'] = 'user4'
            ticket['owner'] = 'user5'
            ticket['summary'] = 'New Ticket Subscribers'
            ticket.insert()
            notify_ticket_created(self.env, ticket)
            return ticket

        def _add_new_ticket_subscriber(sid, authenticated):
            Subscription.add(self.env, {
                'sid': sid, 'authenticated': authenticated,
                'distributor': 'email', 'format': None, 'adverb': 'always',
                'class': 'NewTicketSubscriber'
            })

        _test_new_ticket()
        recipients = smtpd.get_recipients()
        self.assertEqual(0, len(recipients))

        self.env.insert_users([
            ('user1', 'User One', 'joe@example.org', 1),
            ('user2', 'User Two', 'bar@example.org', 1),
            ('58bd3a', 'User Three', 'jim@example.org', 0),
        ])
        _add_new_ticket_subscriber('user1', 1)
        _add_new_ticket_subscriber('58bd3a', 0)

        ticket = _test_new_ticket()
        recipients = smtpd.get_recipients()
        self.assertEqual(2, len(recipients))
        self.assertIn('joe@example.org', recipients)
        self.assertIn('jim@example.org', recipients)

        ticket.save_changes('user4', 'this is my comment',
                            when=datetime_now(utc))
        notify_ticket_changed(self.env, ticket)

        recipients = smtpd.get_recipients()
        self.assertEqual(0, len(recipients))

    def test_no_duplicates(self):
        """Email addresses should be found only once in the recipient list."""
        self.env.config.set('notification', 'smtp_always_cc',
                            'joe.user@example.com')
        ticket = insert_ticket(self.env, reporter='joe.user@example.com',
                               owner='joe.user@example.com',
                               cc='joe.user@example.com',
                               summary='No duplicates')
        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()
        self.assertEqual(1, len(recipients))
        self.assertIn('joe.user@example.com', recipients)

    def test_long_forms(self):
        """Long forms of SMTP email addresses 'Display Name <address>'"""
        config_subscriber(self.env, updater=True, owner=True)
        ticket = insert_ticket(self.env,
           reporter='"Joe" <joe.user@example.com>',
           owner='Joe <joe.user@example.net>',
           cc=u' \u00a0 Jóe \u3000 < joe.user@example.org > \u00a0 ',
           summary='Long form')
        notify_ticket_created(self.env, ticket)
        recipients = smtpd.get_recipients()
        self.assertEqual(3, len(recipients))
        self.assertIn('joe.user@example.com', recipients)
        self.assertIn('joe.user@example.net', recipients)
        self.assertIn('joe.user@example.org', recipients)


class NotificationTestCase(unittest.TestCase):
    """Notification test cases that send email over SMTP"""

    def setUp(self):
        self.env = EnvironmentStub(default_data=True)
        config_smtp(self.env)
        self.env.config.set('trac', 'base_url', 'http://localhost/trac')
        self.env.config.set('project', 'url', 'http://localhost/project.url')
        self.env.config.set('notification', 'smtp_enabled', 'true')
        self.env.config.set('notification', 'smtp_always_cc',
                            'joe.user@example.net, joe.bar@example.net')
        self.env.config.set('notification', 'use_public_cc', 'true')
        self.req = MockRequest(self.env)

    def tearDown(self):
        """Signal the notification test suite that a test is over"""
        smtpd.cleanup()
        self.env.reset_db()

    def _insert_ticket(self, **props):
        reporter = props.pop('reporter', 'joeuser')
        summary = props.pop('summary', 'Summary')
        return insert_ticket(self.env, reporter=reporter, summary=summary,
                             **props)

    def test_structure(self):
        """Basic SMTP message structure (headers, body)"""
        ticket = insert_ticket(self.env,
                               reporter='"Joe User" <joe.user@example.org>',
                               owner='joe.user@example.net',
                               cc='joe.user@example.com, joe.bar@example.org, '
                                  'joe.bar@example.net',
                               summary='This is a summary')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        # checks for header existence
        self.assertTrue(headers)
        # checks for body existence
        self.assertTrue(body)
        # checks for expected headers
        self.assertIn('Date', headers)
        self.assertIn('Subject', headers)
        self.assertEqual('<073.8a48f9c2ab2dc64e820e391f8f784a04@localhost>',
                         headers['Message-ID'])
        self.assertIn('From', headers)
        self.assertIn('\n-- \nTicket URL: <', body)

    def test_date(self):
        """Date format compliance (RFC822)
           we do not support 'military' format"""
        date_str = r"^((?P<day>\w{3}),\s*)*(?P<dm>\d{2})\s+" \
                   r"(?P<month>\w{3})\s+(?P<year>\d{4})\s+" \
                   r"(?P<hour>\d{2}):(?P<min>[0-5][0-9])" \
                   r"(:(?P<sec>[0-5][0-9]))*\s" \
                   r"((?P<tz>\w{2,3})|(?P<offset>[+\-]\d{4}))$"
        date_re = re.compile(date_str)
        # python time module does not detect incorrect time values
        days = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
        months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
        tz = ['UT', 'GMT', 'EST', 'EDT', 'CST', 'CDT', 'MST', 'MDT',
              'PST', 'PDT']
        ticket = insert_ticket(self.env,
                               reporter='"Joe User" <joe.user@example.org>',
                               summary='This is a summary')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertIn('Date', headers)
        mo = date_re.match(headers['Date'])
        self.assertTrue(mo)
        if mo.group('day'):
            self.assertIn(mo.group('day'), days)
        self.assertIn(int(mo.group('dm')), xrange(1, 32))
        self.assertIn(mo.group('month'), months)
        self.assertIn(int(mo.group('hour')), xrange(24))
        if mo.group('tz'):
            self.assertIn(mo.group('tz'), tz)

    def test_bcc_privacy(self):
        """Visibility of recipients"""
        def run_bcc_feature(public_cc):
            # CC list should be private
            self.env.config.set('notification', 'use_public_cc', public_cc)
            self.env.config.set('notification', 'smtp_always_bcc',
                                'joe.foobar@example.net')
            ticket = insert_ticket(self.env,
                                   reporter='"Joe User" <joe.user@example.org>',
                                   summary='This is a summary')
            notify_ticket_created(self.env, ticket)
            message = smtpd.get_message()
            headers, body = parse_smtp_message(message)
            # Msg should have a To header
            self.assertEqual('undisclosed-recipients: ;', headers['To'])
            # Extract the list of 'Cc' recipients from the message
            cc = extract_recipients(headers['Cc'])
            # Extract the list of the actual SMTP recipients
            rcptlist = smtpd.get_recipients()
            # Build the list of the expected 'Cc' recipients
            ccrcpt = self.env.config.getlist('notification', 'smtp_always_cc')
            for rcpt in ccrcpt:
                # Each recipient of the 'Cc' list should appear
                # in the 'Cc' header
                self.assertIn(rcpt, cc)
                # Check the message has actually been sent to the recipients
                self.assertIn(rcpt, rcptlist)
            # Build the list of the expected 'Bcc' recipients
            bccrcpt = self.env.config.getlist('notification',
                                              'smtp_always_bcc')
            for rcpt in bccrcpt:
                # Check the message has actually been sent to the recipients
                self.assertIn(rcpt, rcptlist)
        for public in False, True:
            run_bcc_feature(public)

    def test_short_login(self):
        """Email addresses without a FQDN"""
        def _test_short_login(use_short_addr, username, address):
            config_subscriber(self.env, reporter=True)
            ticket = insert_ticket(self.env, reporter=username,
                                   summary='This is a summary')
            # Be sure that at least one email address is valid, so that we
            # send a notification even if other addresses are not valid
            self.env.config.set('notification', 'smtp_always_cc',
                                'joe.bar@example.net, john')
            self.env.config.set('notification', 'use_short_addr',
                                'enabled' if use_short_addr else 'disabled')
            notify_ticket_created(self.env, ticket)
            message = smtpd.get_message()
            recipients = set(smtpd.get_recipients())
            headers, body = parse_smtp_message(message)
            # Msg should always have a 'To' field
            if use_short_addr:
                self.assertEqual(address, headers['To'])
            else:
                self.assertEqual('undisclosed-recipients: ;', headers['To'])
            # Msg should have a 'Cc' field
            self.assertIn('Cc', headers)
            cclist = set(extract_recipients(headers['Cc']))
            if use_short_addr:
                # Msg should be delivered to the reporter
                self.assertEqual({'joe.bar@example.net', 'john'}, cclist)
                self.assertEqual({address, 'joe.bar@example.net', 'john'},
                                 recipients)
            else:
                # Msg should not be delivered to the reporter
                self.assertEqual({'joe.bar@example.net'}, cclist)
                self.assertEqual({'joe.bar@example.net'}, recipients)

        # Validate with and without the short addr option enabled
        self.env.insert_users([('bar', 'Bar User', ''),
                               ('qux', 'Qux User', 'qux-mail')])
        for use_short_addr in (False, True):
            _test_short_login(use_short_addr, 'foo', 'foo')
            _test_short_login(use_short_addr, 'bar', 'bar')
            _test_short_login(use_short_addr, 'qux', 'qux-mail')

    def test_default_domain(self):
        """Default domain name"""
        def _test_default_domain(enable):
            config_subscriber(self.env)
            self.env.config.set('notification', 'smtp_always_cc', '')
            ticket = insert_ticket(self.env, cc='joenodom, foo, bar, qux, '
                                                'joewithdom@example.com',
                                   summary='This is a summary')
            # Be sure that at least one email address is valid, so that we
            # send a notification even if other addresses are not valid
            self.env.config.set('notification', 'smtp_always_cc',
                                'joe.bar@example.net')
            self.env.config.set('notification', 'smtp_default_domain',
                                'example.org' if enable else '')
            notify_ticket_created(self.env, ticket)
            message = smtpd.get_message()
            headers, body = parse_smtp_message(message)
            # Msg should always have a 'Cc' field
            self.assertIn('Cc', headers)
            cclist = set(extract_recipients(headers['Cc']))
            if enable:
                self.assertEqual({'joenodom@example.org', 'foo@example.org',
                                  'bar@example.org', 'qux-mail@example.org',
                                  'joewithdom@example.com',
                                  'joe.bar@example.net'}, cclist)
            else:
                self.assertEqual({'joewithdom@example.com',
                                  'joe.bar@example.net'}, cclist)

        # Validate with and without a default domain
        self.env.insert_users([('bar', 'Bar User', ''),
                               ('qux', 'Qux User', 'qux-mail')])
        for enable in (False, True):
            _test_default_domain(enable)

    def test_email_map(self):
        """Login-to-email mapping"""
        config_subscriber(self.env, reporter=True, owner=True)
        self.env.config.set('notification', 'smtp_always_cc',
                            'joe@example.com')
        self.env.insert_users(
            [('joeuser', 'Joe User', 'user-joe@example.com'),
             ('jim@domain', 'Jim User', 'user-jim@example.com')])
        ticket = insert_ticket(self.env, reporter='joeuser', owner='jim@domain',
                               summary='This is a summary')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        tolist = sorted(extract_recipients(headers['To']))
        cclist = sorted(extract_recipients(headers['Cc']))
        # 'To' list should have been resolved to the real email address
        self.assertEqual(['user-jim@example.com', 'user-joe@example.com'],
                         tolist)
        self.assertEqual(['joe@example.com'], cclist)

    def test_from_author(self):
        """Using the reporter or change author as the notification sender"""
        self.env.config.set('notification', 'smtp_from', 'trac@example.com')
        self.env.config.set('notification', 'smtp_from_name', 'My Trac')
        self.env.config.set('notification', 'smtp_from_author', 'true')
        self.env.insert_users(
            [('joeuser', 'Joe User', 'user-joe@example.com'),
             ('jim@domain', 'Jim User', 'user-jim@example.com'),
             ('noemail', 'No e-mail', ''),
             ('noname', '', 'user-noname@example.com')])
        def modtime(delta):
            return datetime(2016, 8, 21, 12, 34, 56, 987654, utc) + \
                   timedelta(seconds=delta)
        # Ticket creation uses the reporter
        ticket = insert_ticket(self.env, reporter='joeuser',
                               summary='This is a summary', when=modtime(0))
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('"Joe User" <user-joe@example.com>', headers['From'])
        self.assertEqual('<047.54e62c60198a043f858f1311784a5791@example.com>',
                         headers['Message-ID'])
        self.assertNotIn('In-Reply-To', headers)
        self.assertNotIn('References', headers)
        # Ticket change uses the change author
        ticket['summary'] = 'Modified summary'
        ticket.save_changes('jim@domain', 'Made some changes', modtime(1))
        notify_ticket_changed(self.env, ticket, 'jim@domain')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('"Jim User" <user-jim@example.com>', headers['From'])
        self.assertEqual('<062.a890ee4ad5488fb49e60b68099995ba3@example.com>',
                         headers['Message-ID'])
        self.assertEqual('<047.54e62c60198a043f858f1311784a5791@example.com>',
                         headers['In-Reply-To'])
        self.assertEqual('<047.54e62c60198a043f858f1311784a5791@example.com>',
                         headers['References'])
        # Known author without name uses e-mail address only
        ticket['summary'] = 'Final summary'
        ticket.save_changes('noname', 'Final changes', modtime(2))
        notify_ticket_changed(self.env, ticket, 'noname')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('user-noname@example.com', headers['From'])
        self.assertEqual('<062.732a7b25a21f5a86478c4fe47e86ade4@example.com>',
                         headers['Message-ID'])
        # Known author without e-mail uses smtp_from and smtp_from_name
        ticket['summary'] = 'Other summary'
        ticket.save_changes('noemail', 'More changes', modtime(3))
        notify_ticket_changed(self.env, ticket, 'noemail')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('"My Trac" <trac@example.com>', headers['From'])
        self.assertEqual('<062.98cff27cb9fabd799bcb09f9edd6c99e@example.com>',
                         headers['Message-ID'])
        # Unknown author with name and e-mail address
        ticket['summary'] = 'Some summary'
        ticket.save_changes('Test User <test@example.com>', 'Some changes',
                            modtime(4))
        notify_ticket_changed(self.env, ticket, 'Test User <test@example.com>')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('"Test User" <test@example.com>', headers['From'])
        self.assertEqual('<062.6e08a363c340c1d4e2ed84c6123a1e9d@example.com>',
                         headers['Message-ID'])
        # Unknown author with e-mail address only
        ticket['summary'] = 'Some summary'
        ticket.save_changes('test@example.com', 'Some changes', modtime(5))
        notify_ticket_changed(self.env, ticket, 'test@example.com')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('test@example.com', headers['From'])
        self.assertEqual('<062.5eb050ae6f322c33dde5940a5f318343@example.com>',
                         headers['Message-ID'])
        # Unknown author uses smtp_from and smtp_from_name
        ticket['summary'] = 'Better summary'
        ticket.save_changes('unknown', 'Made more changes', modtime(6))
        notify_ticket_changed(self.env, ticket, 'unknown')
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('"My Trac" <trac@example.com>', headers['From'])
        self.assertEqual('<062.6d5543782e7aba4100302487e75ce16f@example.com>',
                         headers['Message-ID'])

    def test_ignore_domains(self):
        """Non-SMTP domain exclusion"""
        config_subscriber(self.env, reporter=True, owner=True)
        self.env.config.set('notification', 'ignore_domains',
                            'example.com, example.org')
        self.env.insert_users(
            [('kerberos@example.com', 'No Email', ''),
             ('kerberos@example.org', 'With Email', 'kerb@example.net')])
        ticket = insert_ticket(self.env, reporter='kerberos@example.com',
                               owner='kerberos@example.org',
                               summary='This is a summary')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        tolist = set(extract_recipients(headers['To']))
        cclist = set(extract_recipients(headers['Cc']))
        # 'To' list should not contain addresses with non-SMTP domains
        self.assertNotIn('kerberos@example.com', tolist)
        self.assertNotIn('kerberos@example.org', tolist)
        # 'To' list should have been resolved to the actual email address
        self.assertEqual({'kerb@example.net'}, tolist)
        # 'Cc' list should have been resolved to the actual email address
        self.assertEqual({'joe.bar@example.net', 'joe.user@example.net'},
                         cclist)

    def test_admit_domains(self):
        """SMTP domain inclusion"""
        config_subscriber(self.env, reporter=True)
        self.env.config.set('notification', 'admit_domains',
                            'localdomain, server')
        ticket = insert_ticket(self.env, reporter='joeuser@example.com',
                               summary='This is a summary',
                               cc='joe.user@localdomain, joe.user@unknown, '
                                  'joe.user@server')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        # Msg should always have a 'To' field
        self.assertEqual('joeuser@example.com', headers['To'])
        self.assertIn('Cc', headers)
        cclist = set(extract_recipients(headers['Cc']))
        # 'Cc' list should contain addresses with SMTP included domains
        self.assertIn('joe.user@localdomain', cclist)
        self.assertIn('joe.user@server', cclist)
        # 'Cc' list should not contain non-FQDN domains
        self.assertNotIn('joe.user@unknown', cclist)
        self.assertEqual({'joe.user@localdomain', 'joe.user@server',
                          'joe.user@example.net', 'joe.bar@example.net'},
                         cclist)

    def test_multiline_header(self):
        """Encoded headers split into multiple lines"""
        self.env.config.set('notification', 'mime_encoding', 'qp')
        ticket = insert_ticket(self.env, reporter='joe.user@example.org',
                               summary=u'A_very %s súmmäry'
                                       % u' '.join(['long'] * 20))
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        # Discards the project name & ticket number
        subject = headers['Subject']
        summary = subject[subject.find(':')+2:]
        self.assertEqual(ticket['summary'], summary)

    def test_mimebody_b64(self):
        """MIME Base64/utf-8 encoding"""
        self.env.config.set('notification', 'mime_encoding', 'base64')
        summary = u'This is a long enough summary to cause Trac ' \
                  u'to generate a multi-line (2 lines) súmmäry'
        ticket = insert_ticket(self.env, reporter='joe.user@example.org',
                               summary=summary)
        self._validate_mimebody((base64, 'base64', 'utf-8'), ticket, True)

    def test_mimebody_qp(self):
        """MIME QP/utf-8 encoding"""
        self.env.config.set('notification', 'mime_encoding', 'qp')
        summary = u'This is a long enough summary to cause Trac ' \
                  u'to generate a multi-line (2 lines) súmmäry'
        ticket = insert_ticket(self.env, reporter='joe.user@example.org',
                               summary=summary)
        self._validate_mimebody((quopri, 'quoted-printable', 'utf-8'),
                                ticket, True)

    def test_mimebody_none_7bit(self):
        """MIME None encoding resulting in 7bit"""
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = insert_ticket(self.env, reporter='joe.user',
                               summary=u'This is a summary')
        self._validate_mimebody((None, '7bit', 'utf-8'), ticket, True)

    def test_mimebody_none_8bit(self):
        """MIME None encoding resulting in 8bit"""
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = insert_ticket(self.env, reporter='joe.user',
                               summary=u'This is a summary for Jöe Usèr')
        self._validate_mimebody((None, '8bit', 'utf-8'), ticket, True)

    def _test_msgid_digest(self, hash_type):
        """MD5 digest w/ non-ASCII recipient address (#3491)"""
        config_subscriber(self.env, reporter=True)
        self.env.config.set('notification', 'smtp_always_cc', '')
        if hash_type:
            self.env.config.set('notification', 'message_id_hash', hash_type)
        ticket = insert_ticket(self.env, summary=u'This is a summary',
                               reporter=u'"Jöe Usèr" <joe.user@example.org>')
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertEqual('joe.user@example.org', headers['To'])
        self.assertNotIn('Cc', headers)
        return headers

    def test_md5_digest(self):
        headers = self._test_msgid_digest(None)
        self.assertEqual('<071.cbea352f8c4fa58e4b10d24c17b091e6@localhost>',
                         headers['Message-ID'])

    def test_sha1_digest(self):
        headers = self._test_msgid_digest('sha1')
        self.assertEqual(
            '<071.0b6459808bc3603bd642b9a478928d9b5542a803@localhost>',
            headers['Message-ID'])

    def test_add_to_cc_list(self):
        """Members added to CC list receive notifications."""
        config_subscriber(self.env)
        ticket = insert_ticket(self.env, summary='Foo')
        ticket['cc'] = 'joe.user1@example.net'
        now = datetime_now(utc)
        ticket.save_changes('joe.bar@example.com', 'Added to cc', now)
        notify_ticket_changed(self.env, ticket)
        recipients = smtpd.get_recipients()
        self.assertIn('joe.user1@example.net', recipients)

    def test_previous_cc_list(self):
        """Members removed from CC list receive notifications"""
        config_subscriber(self.env)
        ticket = insert_ticket(self.env, summary='Foo',
                               cc='joe.user1@example.net')
        ticket['cc'] = 'joe.user2@example.net'
        now = datetime_now(utc)
        ticket.save_changes('joe.bar@example.com', 'Removed from cc', now)
        notify_ticket_changed(self.env, ticket)
        recipients = smtpd.get_recipients()
        self.assertIn('joe.user1@example.net', recipients)
        self.assertIn('joe.user2@example.net', recipients)

    def test_previous_owner(self):
        """Previous owner is notified when ticket is reassigned (#2311)
           if always_notify_owner is set to True"""
        def _test_owner(enabled):
            config_subscriber(self.env, owner=enabled)
            prev_owner = 'joe.user1@example.net'
            ticket = insert_ticket(self.env, summary='Foo', owner=prev_owner)
            ticket['owner'] = new_owner = 'joe.user2@example.net'
            now = datetime_now(utc)
            ticket.save_changes('joe.bar@example.com', 'Changed owner', now)
            notify_ticket_changed(self.env, ticket)
            recipients = smtpd.get_recipients()
            if enabled:
                self.assertIn(prev_owner, recipients)
                self.assertIn(new_owner, recipients)
            else:
                self.assertNotIn(prev_owner, recipients)
                self.assertNotIn(new_owner, recipients)

        for enable in False, True:
            _test_owner(enable)

    def _validate_mimebody(self, mime, ticket, newtk):
        """Body of a ticket notification message"""
        mime_decoder, mime_name, mime_charset = mime
        if newtk:
            notify_ticket_created(self.env, ticket)
        else:
            notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertIn('MIME-Version', headers)
        self.assertIn('Content-Type', headers)
        self.assertTrue(re.compile(r"1.\d").match(headers['MIME-Version']))
        ctype_mo = re.match(r'\s*([^;\s]*)\s*(?:;\s*boundary="([^"]*)")?',
                            headers['Content-Type'])
        self.assertEqual('multipart/related', ctype_mo.group(1))
        boundary_re = re.compile(r'(?:\r\n)*^--%s(?:--)?(?:\r\n|\Z)' %
                                 re.escape(ctype_mo.group(2)), re.MULTILINE)
        body = boundary_re.split(message)[1]
        headers, body = parse_smtp_message(body)
        self.assertIn('Content-Type', headers)
        self.assertIn('Content-Transfer-Encoding', headers)
        type_re = re.compile(r'^text/plain;\scharset="([\w\-\d]+)"$')
        charset = type_re.match(headers['Content-Type'])
        self.assertTrue(charset)
        charset = charset.group(1)
        self.assertEqual(mime_charset, charset)
        self.assertEqual(headers['Content-Transfer-Encoding'], mime_name)
        # checks the width of each body line
        for line in body.splitlines():
            self.assertTrue(len(line) <= MAXBODYWIDTH)
        # attempts to decode the body, following the specified MIME encoding
        # and charset
        try:
            if mime_decoder:
                body = mime_decoder.decodestring(body)
            body = unicode(body, charset)
        except Exception as e:
            raise AssertionError(e)
        # now processes each line of the body
        bodylines = body.splitlines()
        # body starts with one of more summary lines, first line is prefixed
        # with the ticket number such as #<n>: summary
        # finds the banner after the summary
        banner_delim_re = re.compile(r'^\-+\+\-+$')
        bodyheader = []
        while not banner_delim_re.match(bodylines[0]):
            bodyheader.append(bodylines.pop(0))
        # summary should be present
        self.assertTrue(bodyheader)
        # banner should not be empty
        self.assertTrue(bodylines)
        # extracts the ticket ID from the first line
        tknum, bodyheader[0] = bodyheader[0].split(' ', 1)
        self.assertEqual('#', tknum[0])
        try:
            tkid = int(tknum[1:-1])
            self.assertEqual(1, tkid)
        except ValueError:
            raise AssertionError("invalid ticket number")
        self.assertEqual(':', tknum[-1])
        summary = ' '.join(bodyheader)
        self.assertEqual(summary, ticket['summary'])
        # now checks the banner contents
        self.assertTrue(banner_delim_re.match(bodylines[0]))
        banner = True
        footer = None
        props = {}
        for line in bodylines[1:]:
            # detect end of banner
            if banner_delim_re.match(line):
                banner = False
                continue
            if banner:
                # parse banner and fill in a property dict
                properties = line.split('|')
                self.assertEqual(2, len(properties))
                for prop in properties:
                    if prop.strip() == '':
                        continue
                    k, v = prop.split(':')
                    props[k.strip().lower()] = v.strip()
            # detect footer marker (weak detection)
            if not footer:
                if line.strip() == '--':
                    footer = 0
                    continue
            # check footer
            if footer is not None:
                footer += 1
                # invalid footer detection
                self.assertTrue(footer <= 3)
                # check ticket link
                if line[:11] == 'Ticket URL:':
                    ticket_link = self.env.abs_href.ticket(ticket.id)
                    self.assertEqual(line[12:].strip(), "<%s>" % ticket_link)
                # note project title / URL are not validated yet

        # ticket properties which are not expected in the banner
        xlist = ['summary', 'description', 'comment', 'time', 'changetime']
        # check banner content (field exists, msg value matches ticket value)
        for p in [prop for prop in ticket.values if prop not in xlist]:
            self.assertIn(p, props)
            # Email addresses might be obfuscated
            if '@' in ticket[p] and '@' in props[p]:
                self.assertEqual(props[p].split('@')[0],
                                 ticket[p].split('@')[0])
            else:
                self.assertEqual(props[p], ticket[p])

    def test_props_format_ambiwidth_single(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        self.env.config.set('notification', 'ambiguous_char_width', '')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'аnonymoиs'
        ticket['status'] = u'new'
        ticket['owner'] = u'somеbody'
        ticket['type'] = u'バグ(dеfеct)'
        ticket['priority'] = u'メジャー(mаjor)'
        ticket['milestone'] = u'マイルストーン1'
        ticket['component'] = u'コンポーネント1'
        ticket['version'] = u'2.0 аlphа'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  аnonymoиs        |      Owner:  somеbody
      Type:  バグ(dеfеct)     |     Status:  new
  Priority:  メジャー(mаjor)  |  Milestone:  マイルストーン1
 Component:  コンポーネント1  |    Version:  2.0 аlphа
Resolution:  fixed            |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_ambiwidth_double(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        self.env.config.set('notification', 'ambiguous_char_width', 'double')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'аnonymoиs'
        ticket['status'] = u'new'
        ticket['owner'] = u'somеbody'
        ticket['type'] = u'バグ(dеfеct)'
        ticket['priority'] = u'メジャー(mаjor)'
        ticket['milestone'] = u'マイルストーン1'
        ticket['component'] = u'コンポーネント1'
        ticket['version'] = u'2.0 аlphа'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  аnonymoиs       |      Owner:  somеbody
      Type:  バグ(dеfеct)    |     Status:  new
  Priority:  メジャー(mаjor)  |  Milestone:  マイルストーン1
 Component:  コンポーネント1   |    Version:  2.0 аlphа
Resolution:  fixed             |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_obfuscated_email(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'joe@foobar.foo.bar.example.org'
        ticket['status'] = u'new'
        ticket['owner'] = u'joe.bar@foobar.foo.bar.example.org'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'milestone1'
        ticket['component'] = u'component1'
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  joe@…       |      Owner:  joe.bar@…
      Type:  defect      |     Status:  new
  Priority:  major       |  Milestone:  milestone1
 Component:  component1  |    Version:  2.0
Resolution:  fixed       |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_obfuscated_email_disabled(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        self.env.config.set('trac', 'show_email_addresses', 'true')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'joe@foobar.foo.bar.example.org'
        ticket['status'] = u'new'
        ticket['owner'] = u'joe.bar@foobar.foo.bar.example.org'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'milestone1'
        ticket['component'] = u'component1'
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:                          |      Owner:
  joe@foobar.foo.bar.example.org     |  joe.bar@foobar.foo.bar.example.org
      Type:  defect                  |     Status:  new
  Priority:  major                   |  Milestone:  milestone1
 Component:  component1              |    Version:  2.0
Resolution:  fixed                   |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_show_full_names(self):
        self.env.insert_users([
            ('joefoo', u'Joę Fœœ', 'joe@foobar.foo.bar.example.org'),
            ('joebar', u'Jœe Bær', 'joe.bar@foobar.foo.bar.example.org')
        ])
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = 'This is a summary'
        ticket['reporter'] = 'joefoo'
        ticket['status'] = 'new'
        ticket['owner'] = 'joebar'
        ticket['type'] = 'defect'
        ticket['priority'] = 'major'
        ticket['milestone'] = 'milestone1'
        ticket['component'] = 'component1'
        ticket['version'] = '2.0'
        ticket['resolution'] = 'fixed'
        ticket['keywords'] = ''
        ticket.insert()
        formatted = """\
  Reporter:  Joę Fœœ     |      Owner:  Jœe Bær
      Type:  defect      |     Status:  new
  Priority:  major       |  Milestone:  milestone1
 Component:  component1  |    Version:  2.0
Resolution:  fixed       |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_leftside(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'milestone1'
        ticket['component'] = u'Lorem ipsum dolor sit amet, consectetur ' \
                              u'adipisicing elit, sed do eiusmod tempor ' \
                              u'incididunt ut labore et dolore magna ' \
                              u'aliqua. Ut enim ad minim veniam, quis ' \
                              u'nostrud exercitation ullamco laboris nisi ' \
                              u'ut aliquip ex ea commodo consequat. Duis ' \
                              u'aute irure dolor in reprehenderit in ' \
                              u'voluptate velit esse cillum dolore eu ' \
                              u'fugiat nulla pariatur. Excepteur sint ' \
                              u'occaecat cupidatat non proident, sunt in ' \
                              u'culpa qui officia deserunt mollit anim id ' \
                              u'est laborum.'
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  anonymous                           |      Owner:  somebody
      Type:  defect                              |     Status:  new
  Priority:  major                               |  Milestone:  milestone1
 Component:  Lorem ipsum dolor sit amet,         |    Version:  2.0
  consectetur adipisicing elit, sed do eiusmod   |
  tempor incididunt ut labore et dolore magna    |
  aliqua. Ut enim ad minim veniam, quis nostrud  |
  exercitation ullamco laboris nisi ut aliquip   |
  ex ea commodo consequat. Duis aute irure       |
  dolor in reprehenderit in voluptate velit      |
  esse cillum dolore eu fugiat nulla pariatur.   |
  Excepteur sint occaecat cupidatat non          |
  proident, sunt in culpa qui officia deserunt   |
  mollit anim id est laborum.                    |
Resolution:  fixed                               |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_leftside_unicode(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'milestone1'
        ticket['component'] = u'Trac は BSD ライセンスのもとで配' \
                              u'布されています。[1:]このライセ' \
                              u'ンスの全文は、配布ファイルに' \
                              u'含まれている [3:COPYING] ファイル' \
                              u'と同じものが[2:オンライン]で参' \
                              u'照できます。'
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  anonymous                           |      Owner:  somebody
      Type:  defect                              |     Status:  new
  Priority:  major                               |  Milestone:  milestone1
 Component:  Trac は BSD ライセンスのもとで配布  |    Version:  2.0
  されています。[1:]このライセンスの全文は、配   |
  布ファイルに含まれている [3:COPYING] ファイル  |
  と同じものが[2:オンライン]で参照できます。     |
Resolution:  fixed                               |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_rightside(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'Lorem ipsum dolor sit amet, consectetur ' \
                              u'adipisicing elit, sed do eiusmod tempor ' \
                              u'incididunt ut labore et dolore magna ' \
                              u'aliqua. Ut enim ad minim veniam, quis ' \
                              u'nostrud exercitation ullamco laboris nisi ' \
                              u'ut aliquip ex ea commodo consequat. Duis ' \
                              u'aute irure dolor in reprehenderit in ' \
                              u'voluptate velit esse cillum dolore eu ' \
                              u'fugiat nulla pariatur. Excepteur sint ' \
                              u'occaecat cupidatat non proident, sunt in ' \
                              u'culpa qui officia deserunt mollit anim id ' \
                              u'est laborum.'
        ticket['component'] = u'component1'
        ticket['version'] = u'2.0 Standard and International Edition'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  anonymous   |      Owner:  somebody
      Type:  defect      |     Status:  new
  Priority:  major       |  Milestone:  Lorem ipsum dolor sit amet,
                         |  consectetur adipisicing elit, sed do eiusmod
                         |  tempor incididunt ut labore et dolore magna
                         |  aliqua. Ut enim ad minim veniam, quis nostrud
                         |  exercitation ullamco laboris nisi ut aliquip ex
                         |  ea commodo consequat. Duis aute irure dolor in
                         |  reprehenderit in voluptate velit esse cillum
                         |  dolore eu fugiat nulla pariatur. Excepteur sint
                         |  occaecat cupidatat non proident, sunt in culpa
                         |  qui officia deserunt mollit anim id est
                         |  laborum.
 Component:  component1  |    Version:  2.0 Standard and International
                         |  Edition
Resolution:  fixed       |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_rightside_unicode(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'Trac 在经过修改的BSD协议下发布。' \
                              u'[1:]协议的完整文本可以[2:在线查' \
                              u'看]也可在发布版的 [3:COPYING] 文' \
                              u'件中找到。'
        ticket['component'] = u'component1'
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  anonymous   |      Owner:  somebody
      Type:  defect      |     Status:  new
  Priority:  major       |  Milestone:  Trac 在经过修改的BSD协议下发布。
                         |  [1:]协议的完整文本可以[2:在线查看]也可在发布版
                         |  的 [3:COPYING] 文件中找到。
 Component:  component1  |    Version:  2.0
Resolution:  fixed       |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_bothsides(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'Lorem ipsum dolor sit amet, consectetur ' \
                              u'adipisicing elit, sed do eiusmod tempor ' \
                              u'incididunt ut labore et dolore magna ' \
                              u'aliqua. Ut enim ad minim veniam, quis ' \
                              u'nostrud exercitation ullamco laboris nisi ' \
                              u'ut aliquip ex ea commodo consequat. Duis ' \
                              u'aute irure dolor in reprehenderit in ' \
                              u'voluptate velit esse cillum dolore eu ' \
                              u'fugiat nulla pariatur. Excepteur sint ' \
                              u'occaecat cupidatat non proident, sunt in ' \
                              u'culpa qui officia deserunt mollit anim id ' \
                              u'est laborum.'
        ticket['component'] = (u'Lorem ipsum dolor sit amet, consectetur '
                               u'adipisicing elit, sed do eiusmod tempor '
                               u'incididunt ut labore et dolore magna aliqua.')
        ticket['version'] = u'2.0'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u'Ut enim ad minim veniam, ....'
        ticket.insert()
        formatted = """\
  Reporter:  anonymous               |      Owner:  somebody
      Type:  defect                  |     Status:  new
  Priority:  major                   |  Milestone:  Lorem ipsum dolor sit
                                     |  amet, consectetur adipisicing elit,
                                     |  sed do eiusmod tempor incididunt ut
                                     |  labore et dolore magna aliqua. Ut
                                     |  enim ad minim veniam, quis nostrud
                                     |  exercitation ullamco laboris nisi
                                     |  ut aliquip ex ea commodo consequat.
                                     |  Duis aute irure dolor in
                                     |  reprehenderit in voluptate velit
                                     |  esse cillum dolore eu fugiat nulla
 Component:  Lorem ipsum dolor sit   |  pariatur. Excepteur sint occaecat
  amet, consectetur adipisicing      |  cupidatat non proident, sunt in
  elit, sed do eiusmod tempor        |  culpa qui officia deserunt mollit
  incididunt ut labore et dolore     |  anim id est laborum.
  magna aliqua.                      |    Version:  2.0
Resolution:  fixed                   |   Keywords:  Ut enim ad minim
                                     |  veniam, ...."""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_bothsides_unicode(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        self.env.config.set('notification', 'ambiguous_char_width', 'double')
        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['status'] = u'new'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['priority'] = u'major'
        ticket['milestone'] = u'Trac 在经过修改的BSD协议下发布。' \
                              u'[1:]协议的完整文本可以[2:在线查' \
                              u'看]也可在发布版的 [3:COPYING] 文' \
                              u'件中找到。'
        ticket['component'] = u'Trac は BSD ライセンスのもとで配' \
                              u'布されています。[1:]このライセ' \
                              u'ンスの全文は、※配布ファイル' \
                              u'に含まれている[3:CОPYING]ファイ' \
                              u'ルと同じものが[2:オンライン]で' \
                              u'参照できます。'
        ticket['version'] = u'2.0 International Edition'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket.insert()
        formatted = """\
  Reporter:  anonymous               |      Owner:  somebody
      Type:  defect                  |     Status:  new
  Priority:  major                   |  Milestone:  Trac 在经过修改的BSD协
 Component:  Trac は BSD ライセンス  |  议下发布。[1:]协议的完整文本可以[2:
  のもとで配布されています。[1:]こ   |  在线查看]也可在发布版的 [3:COPYING]
  のライセンスの全文は、※配布ファ   |  文件中找到。
  イルに含まれている[3:CОPYING]フ   |    Version:  2.0 International
  ァイルと同じものが[2:オンライン]   |  Edition
  で参照できます。                   |
Resolution:  fixed                   |   Keywords:"""
        self._validate_props_format(formatted, ticket)

    def test_props_format_wrap_ticket_10283(self):
        self.env.config.set('notification', 'mime_encoding', 'none')
        for name, value in (('blockedby', 'text'),
                            ('blockedby.label', 'Blocked by'),
                            ('blockedby.order', '6'),
                            ('blocking', 'text'),
                            ('blocking.label', 'Blocking'),
                            ('blocking.order', '5'),
                            ('deployment', 'text'),
                            ('deployment.label', 'Deployment state'),
                            ('deployment.order', '1'),
                            ('nodes', 'text'),
                            ('nodes.label', 'Related nodes'),
                            ('nodes.order', '3'),
                            ('privacy', 'text'),
                            ('privacy.label', 'Privacy sensitive'),
                            ('privacy.order', '2'),
                            ('sensitive', 'text'),
                            ('sensitive.label', 'Security sensitive'),
                            ('sensitive.order', '4')):
            self.env.config.set('ticket-custom', name, value)

        ticket = Ticket(self.env)
        ticket['summary'] = u'This is a summary'
        ticket['reporter'] = u'anonymous'
        ticket['owner'] = u'somebody'
        ticket['type'] = u'defect'
        ticket['status'] = u'closed'
        ticket['priority'] = u'normal'
        ticket['milestone'] = u'iter_01'
        ticket['component'] = u'XXXXXXXXXXXXXXXXXXXXXXXXXX'
        ticket['resolution'] = u'fixed'
        ticket['keywords'] = u''
        ticket['deployment'] = ''
        ticket['privacy'] = '0'
        ticket['nodes'] = 'XXXXXXXXXX'
        ticket['sensitive'] = '0'
        ticket['blocking'] = ''
        ticket['blockedby'] = ''
        ticket.insert()

        formatted = """\
          Reporter:  anonymous                   |             Owner:
                                                 |  somebody
              Type:  defect                      |            Status:
                                                 |  closed
          Priority:  normal                      |         Milestone:
                                                 |  iter_01
         Component:  XXXXXXXXXXXXXXXXXXXXXXXXXX  |        Resolution:
                                                 |  fixed
          Keywords:                              |  Deployment state:
 Privacy sensitive:  0                           |     Related nodes:
                                                 |  XXXXXXXXXX
Security sensitive:  0                           |          Blocking:
        Blocked by:                              |"""
        self._validate_props_format(formatted, ticket)

    def _validate_props_format(self, expected, ticket):
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        bodylines = body.splitlines()
        # Extract ticket properties
        delim_re = re.compile(r'^\-+\+\-+$')
        while not delim_re.match(bodylines[0]):
            bodylines.pop(0)
        lines = []
        for line in bodylines[1:]:
            if delim_re.match(line):
                break
            lines.append(line)
        self.assertEqual(expected, '\n'.join(lines))

    def test_notification_does_not_alter_ticket_instance(self):
        ticket = insert_ticket(self.env, summary='My Summary',
                               description='Some description')
        notify_ticket_created(self.env, ticket)
        self.assertIsNotNone(smtpd.get_message())
        self.assertEqual('My Summary', ticket['summary'])
        self.assertEqual('Some description', ticket['description'])
        valid_fieldnames = {f['name'] for f in ticket.fields}
        current_fieldnames = set(ticket.values)
        self.assertEqual(set(), current_fieldnames - valid_fieldnames)

    def test_mime_meta_characters_in_from_header(self):
        """MIME encoding with meta characters in From header"""

        self.env.config.set('notification', 'smtp_from', 'trac@example.com')
        self.env.config.set('notification', 'mime_encoding', 'base64')
        ticket = insert_ticket(self.env, reporter='joeuser',
                               summary='This is a summary')

        def notify(from_name):
            self.env.config.set('notification', 'smtp_from_name', from_name)
            notify_ticket_created(self.env, ticket)
            message = smtpd.get_message()
            headers, body = parse_smtp_message(message)
            return message, headers, body

        message, headers, body = notify(u'Träc')
        self.assertEqual(r'"=?utf-8?b?VHLDpGM=?=" <trac@example.com>',
                         headers['From'])
        message, headers, body = notify(u'Trac\\')
        self.assertEqual(r'"Trac\\" <trac@example.com>', headers['From'])
        message, headers, body = notify(u'Trac"')
        self.assertEqual(r'"Trac\"" <trac@example.com>', headers['From'])
        message, headers, body = notify(u'=?utf-8?b?****?=')
        self.assertEqual('"=?utf-8?b?PT91dGYtOD9iPyoqKio/PQ==?=" '
                         '<trac@example.com>', headers['From'])

    def test_mime_meta_characters_in_subject_header(self):
        """MIME encoding with meta characters in Subject header"""

        self.env.config.set('notification', 'smtp_from', 'trac@example.com')
        self.env.config.set('notification', 'mime_encoding', 'base64')
        summary = u'=?utf-8?q?****?='
        ticket = insert_ticket(self.env, reporter='joeuser', summary=summary)
        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        self.assertIn('\nSubject: =?utf-8?b?', message)  # is mime-encoded
        self.assertEqual(summary,
                         re.split(r' #[0-9]+: ', headers['Subject'], 1)[1])

    def test_mail_headers(self):
        def validates(headers):
            self.assertEqual('http://localhost/project.url',
                             headers.get('X-URL'))
            self.assertEqual('ticket', headers.get('X-Trac-Realm'))
            self.assertEqual(str(ticket.id), headers.get('X-Trac-Ticket-ID'))

        when = datetime(2015, 1, 1, tzinfo=utc)
        ticket = insert_ticket(self.env, reporter='joeuser', summary='Summary',
                               when=when)
        notify_ticket_created(self.env, ticket)
        headers, body = parse_smtp_message(smtpd.get_message())
        validates(headers)
        self.assertEqual('http://localhost/trac/ticket/%d' % ticket.id,
                         headers.get('X-Trac-Ticket-URL'))

        ticket.save_changes(comment='New comment 1',
                            when=when + timedelta(days=1))
        notify_ticket_changed(self.env, ticket)
        headers, body = parse_smtp_message(smtpd.get_message())
        validates(headers)
        self.assertEqual('http://localhost/trac/ticket/%d#comment:1' %
                         ticket.id, headers.get('X-Trac-Ticket-URL'))

        ticket.save_changes(comment='Reply to comment:1', replyto='1',
                            when=when + timedelta(days=2))
        notify_ticket_changed(self.env, ticket)
        headers, body = parse_smtp_message(smtpd.get_message())
        validates(headers)
        self.assertEqual('http://localhost/trac/ticket/%d#comment:2' %
                         ticket.id, headers.get('X-Trac-Ticket-URL'))

    def test_property_change_author_is_obfuscated(self):
        ticket = self._insert_ticket(owner='user1@d.com',
                                     reporter='user2@d.com',
                                     cc='user3@d.com, user4@d.com')
        ticket['owner'] = 'user2@d.com'
        ticket['reporter'] = 'user1@d.com'
        ticket['cc'] = 'user4@d.com'
        ticket.save_changes('user0@d.com', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by user0@…)', body)
        self.assertIn('* owner:  user1@… => user2@…\n', body)
        self.assertIn('* reporter:  user2@… => user1@…\n', body)
        self.assertIn('* cc: user3@… (removed)\n', body)

    def test_comment_change_author_is_obfuscated(self):
        ticket = self._insert_ticket()
        ticket.save_changes('user@d.com', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Comment (by user@…)', body)

    def test_property_change_author_is_not_obfuscated(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.config.set('trac', 'show_full_names', False)
        ticket = self._insert_ticket(owner='user1@d.com',
                                     reporter='user2@d.com',
                                     cc='user3@d.com, user4@d.com')
        ticket['owner'] = 'user2@d.com'
        ticket['reporter'] = 'user1@d.com'
        ticket['cc'] = 'user4@d.com'
        ticket.save_changes('user0@d.com', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by user0@d.com)', body)
        self.assertIn('* owner:  user1@d.com => user2@d.com\n', body)
        self.assertIn('* reporter:  user2@d.com => user1@d.com\n', body)
        self.assertIn('* cc: user3@d.com (removed)\n', body)

    def test_comment_author_is_not_obfuscated(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.config.set('trac', 'show_full_names', False)
        ticket = self._insert_ticket()
        ticket.save_changes('user@d.com', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Comment (by user@d.com)', body)

    def test_property_change_author_full_name(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.insert_users([
            ('user0', u'Ußęr0', 'user0@d.org'),
            ('user1', u'Ußęr1', 'user1@d.org'),
            ('user2', u'Ußęr2', 'user2@d.org'),
            ('user3', u'Ußęr3', 'user3@d.org'),
            ('user4', u'Ußęr4', 'user4@d.org'),
        ])
        ticket = self._insert_ticket(owner='user1', reporter='user2',
                                     cc='user3, user4')
        ticket['owner'] = 'user2'
        ticket['reporter'] = 'user1'
        ticket['cc'] = 'user4'
        ticket.save_changes('user0', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by Ußęr0)', body)
        self.assertIn('* owner:  Ußęr1 => Ußęr2\n', body)
        self.assertIn('* reporter:  Ußęr2 => Ußęr1\n', body)
        self.assertIn('* cc: Ußęr3 (removed)\n', body)

    def test_comment_author_full_name(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.insert_users([
            ('user', u'Thę Ußęr', 'user@domain.org')
        ])
        ticket = self._insert_ticket()
        ticket.save_changes('user', "The comment")

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Comment (by Thę Ußęr)', body)


class FormatSubjectTestCase(unittest.TestCase):

    custom_template = """\
${prefix} (${
   'new' if not changes
   else ticket.resolution if ticket.status == 'closed'
   else ticket.status if 'status' in changes.fields
   else 'commented' if 'comment' in changes.fields
                       and changes.fields['comment']['new']
   else 'updated'
}) #${ticket.id}: ${summary}"""


    def setUp(self):
        self.env = EnvironmentStub()
        TicketNotificationSystem(self.env).environment_created()
        self.env.config.set('project', 'name', 'TracTest')
        self.env.config.set('notification', 'smtp_port', str(SMTP_TEST_PORT))
        self.env.config.set('notification', 'smtp_server', 'localhost')
        self.env.config.set('notification', 'smtp_enabled', 'true')

    def tearDown(self):
        smtpd.cleanup()
        self.env.reset_db()

    def _insert_ticket(self):
        return insert_ticket(self.env, reporter='user@domain.com',
                             summary='The summary',
                             description='The description')

    def test_format_subject_new_ticket(self):
        ticket = self._insert_ticket()

        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('[TracTest] #1: The summary', headers['Subject'])

    def test_format_subject_ticket_change(self):
        ticket = self._insert_ticket()
        ticket['description'] = 'The changed description'
        ticket.save_changes(author='user@domain.com')

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] #1: The summary', headers['Subject'])

    def test_format_subject_ticket_summary_changed(self):
        ticket = self._insert_ticket()
        ticket['summary'] = 'The changed summary'
        ticket.save_changes(author='user@domain.com')

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] #1: The changed summary '
                         '(was: The summary)', headers['Subject'])

    def test_format_subject_custom_template_new_ticket(self):
        """Format subject with a custom template for a new ticket."""
        ticket = self._insert_ticket()
        self.env.config.set('notification', 'ticket_subject_template',
                            self.custom_template)

        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('[TracTest] (new) #1: The summary',
                         headers['Subject'])

    def test_format_subject_custom_template_changed_ticket(self):
        """Format subject with a custom template for a ticket with
        a changed property.
        """
        ticket = self._insert_ticket()
        ticket['description'] = 'The changed description'
        ticket.save_changes(author='user@domain.com')
        self.env.config.set('notification', 'ticket_subject_template',
                            self.custom_template)

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] (updated) #1: The summary',
                         headers['Subject'])

    def test_format_subject_custom_template_commented_ticket(self):
        """Format subject with a custom template for a ticket with
        a changed property and a comment.
        """
        ticket = self._insert_ticket()
        ticket['description'] = 'The changed description'
        ticket.save_changes(author='user@domain.com', comment='the comment')
        self.env.config.set('notification', 'ticket_subject_template',
                            self.custom_template)

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] (commented) #1: The summary',
                         headers['Subject'])

    def test_format_subject_custom_template_status_changed_ticket(self):
        """Format subject with a custom template for a ticket with
        changed status.
        """
        ticket = self._insert_ticket()
        ticket['status'] = 'accepted'
        ticket.save_changes(author='user@domain.com')
        self.env.config.set('notification', 'ticket_subject_template',
                            self.custom_template)

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] (accepted) #1: The summary',
                         headers['Subject'])

    def test_format_subject_custom_template_closed_ticket(self):
        """Format subject with a custom template for a closed ticket."""
        ticket = self._insert_ticket()
        ticket['status'] = 'closed'
        ticket['resolution'] = 'worksforme'
        ticket.save_changes(author='user@domain.com')
        self.env.config.set('notification', 'ticket_subject_template',
                            self.custom_template)

        notify_ticket_changed(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('Re: [TracTest] (worksforme) #1: The summary',
                         headers['Subject'])

    def test_format_subject_custom_template_with_hash(self):
        """Format subject with a custom template with leading #."""
        ticket = self._insert_ticket()
        self.env.config.set('notification', 'ticket_subject_template',
                            '#${ticket.id}: ${summary}')

        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('#1: The summary', headers['Subject'])

    def test_format_subject_custom_template_with_double_hash(self):
        """Format subject with a custom template with leading ##."""
        ticket = self._insert_ticket()
        self.env.config.set('notification', 'ticket_subject_template',
                            '##${prefix}## #${ticket.id}: ${summary}')

        notify_ticket_created(self.env, ticket)
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertEqual('##[TracTest]## #1: The summary', headers['Subject'])


class AttachmentNotificationTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True, path=mkdtemp())
        config_smtp(self.env)
        config_subscriber(self.env, reporter=True)

    def tearDown(self):
        """Signal the notification test suite that a test is over"""
        smtpd.cleanup()
        self.env.reset_db_and_disk()

    def _insert_attachment(self, author):
        ticket = insert_ticket(self.env, summary='Ticket summary',
                               reporter=author)
        attachment = Attachment(self.env, 'ticket', ticket.id)
        attachment.description = "The attachment description"
        attachment.author = author
        attachment.insert('foo.txt', io.BytesIO(), 1)
        return attachment

    def test_ticket_notify_attachment_enabled_attachment_added(self):
        self._insert_attachment('user@example.com')

        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertIn("Re: [TracTest] #1: Ticket summary", headers['Subject'])
        self.assertIn(" * Attachment \"foo.txt\" added", body)
        self.assertIn("The attachment description", body)

    def test_ticket_notify_attachment_enabled_attachment_removed(self):
        attachment = self._insert_attachment('user@example.com')
        attachment.delete()

        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)

        self.assertIn("Re: [TracTest] #1: Ticket summary", headers['Subject'])
        self.assertIn(" * Attachment \"foo.txt\" removed", body)
        self.assertIn("The attachment description", body)

    def test_author_is_obfuscated(self):
        self.env.config.set('trac', 'show_email_addresses', False)
        self.env.config.set('trac', 'show_full_names', False)
        self._insert_attachment('user@example.com')

        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by user@…)', body)

    def test_author_is_not_obfuscated(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.config.set('trac', 'show_full_names', False)
        self._insert_attachment('user@example.com')

        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by user@example.com)', body)

    def test_author_full_name(self):
        self.env.config.set('trac', 'show_email_addresses', True)
        self.env.insert_users([
            ('user', u'Thę Ußęr', 'user@domain.org')
        ])
        self._insert_attachment('user')

        message = smtpd.get_message()
        body = parse_smtp_message(message)[1]

        self.assertIn('Changes (by Thę Ußęr)', body)


class BatchTicketNotificationTestCase(unittest.TestCase):

    def setUp(self):
        self.env = EnvironmentStub(default_data=True, path=mkdtemp())
        config_smtp(self.env)
        self.env.config.set('project', 'url', 'http://localhost/project.url')

        self.tktids = []
        with self.env.db_transaction as db:
            for n in xrange(2):
                for priority in ('', 'blah', 'blocker', 'critical', 'major',
                                 'minor', 'trivial'):
                    idx = len(self.tktids)
                    owner = 'owner@example.org' if idx == 0 else 'anonymous'
                    reporter = 'reporter@example.org' \
                               if idx == 1 else 'anonymous'
                    cc = 'cc1@example.org, cc2@example.org' if idx == 2 else ''
                    when = datetime(2001, 7, 12, 12, 34, idx, 0, utc)
                    ticket = insert_ticket(self.env, summary='Summary %s:%d'
                                                             % (priority, idx),
                                           priority=priority, owner=owner,
                                           reporter=reporter, cc=cc, when=when)
                    self.tktids.append(ticket.id)
        self.tktids.reverse()
        config_subscriber(self.env, updater=True, reporter=True)

    def tearDown(self):
        smtpd.cleanup()
        self.env.reset_db_and_disk()

    def _change_tickets(self, author, new_values, comment, when=None):
        if when is None:
            when = datetime(2016, 8, 21, 12, 34, 56, 987654, utc)
        with self.env.db_transaction:
            for tktid in self.tktids:
                t = Ticket(self.env, tktid)
                for name, value in new_values.iteritems():
                    t[name] = value
                t.save_changes(author, comment, when=when)
        return BatchTicketChangeEvent(self.tktids, when, author, comment,
                                      new_values, 'leave')
    def _notify(self, event):
        smtpd.cleanup()
        NotificationSystem(self.env).notify(event)
        recipients = sorted(smtpd.get_recipients())
        sender = smtpd.get_sender()
        message = smtpd.get_message()
        headers, body = parse_smtp_message(message)
        body = body.splitlines()
        return recipients, sender, message, headers, body

    def test_batchmod_notify(self):
        self.assertEqual(1, min(self.tktids))
        self.assertEqual(14, max(self.tktids))
        new_values = {'milestone': 'milestone1'}
        author = 'author@example.org'
        comment = 'batch-modify'
        when = datetime(2016, 8, 21, 12, 34, 56, 987654, utc)
        event = self._change_tickets(author, new_values, comment, when)

        recipients, sender, message, headers, body = self._notify(event)

        self.assertEqual(['author@example.org', 'cc1@example.org',
                          'cc2@example.org', 'reporter@example.org'],
                         recipients)
        self.assertEqual('trac@localhost', sender)
        self.assertIn('Date', headers)
        self.assertEqual('[TracTest] Batch modify: #3, #10, #4, #11, #5, #12, '
                         '#6, #13, #7, #14,...', headers['Subject'])
        self.assertEqual('"TracTest" <trac@localhost>', headers['From'])
        self.assertEqual('<078.0b9de298f9080302285a0e333c75dd47@localhost>',
                         headers['Message-ID'])
        self.assertIn('Batch modification to #3, #10, #4, #11, #5, #12, #6, '
                      '#13, #7, #14, #1, #2, #8, #9 by author@example.org:',
                      body)
        self.assertIn('-- ', body)
        self.assertIn('Tickets URL: <http://example.org/trac.cgi/query?id=3'
                      '%2C10%2C4%2C11%2C5%2C12%2C6%2C13%2C7%2C14%2C1%2C2%2C8'
                      '%2C9>', body)

    def test_format_subject_custom_template_with_hash(self):
        """Format subject with a custom template with leading #."""
        self.env.config.set('notification', 'batch_subject_template',
                            '#${prefix}# Batch modify')

        event = self._change_tickets(
            author='author@example.org',
            new_values={'milestone': 'milestone1'},
            comment='batch-modify')

        recipients, sender, message, headers, body = self._notify(event)

        self.assertEqual('#[TracTest]# Batch modify', headers['Subject'])

    def test_format_subject_custom_template_with_double_hash(self):
        """Format subject with a custom template with leading ##."""
        self.env.config.set('notification', 'batch_subject_template',
                            '##${prefix}## Batch modify')

        event = self._change_tickets(
            author='author@example.org',
            new_values={'milestone': 'milestone1'},
            comment='batch-modify')

        recipients, sender, message, headers, body = self._notify(event)

        self.assertEqual('##[TracTest]## Batch modify', headers['Subject'])


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RecipientTestCase))
    suite.addTest(unittest.makeSuite(NotificationTestCase))
    suite.addTest(unittest.makeSuite(FormatSubjectTestCase))
    suite.addTest(unittest.makeSuite(AttachmentNotificationTestCase))
    suite.addTest(unittest.makeSuite(BatchTicketNotificationTestCase))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2023 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.org/wiki/TracLicense.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at https://trac.edgewall.org/log/.

import unittest

from trac.search.tests import web_ui
from trac.search.tests.functional import functionalSuite


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(web_ui.test_suite())
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
